!--------------------------------------------------------------------------
MODULE leastsquares_mod

!     Module for unconstrained linear least-squares calculations.
!     The algorithm is suitable for updating LS calculations as more
!     data are added.   This is sometimes called recursive estimation.
!     Only one dependent variable is allowed.
!     Based upon Applied Statistics algorithm AS 274.
!     Translation from Fortran 77 to Fortran 90 by Alan Miller.
!     A subroutine, VARPRD, has been added for calculating the variances
!     of predicted values, and this uses a subroutine BKSUB2.

!     Version 1.01, 5 June 1995
!     Author: Alan Miller
!             CSIRO Division of Mathematics & Statistics
!             Private Bag 10, Rosebank MDC
!             Clayton 3169, Victoria, Australia
!     Phone: (+61) 3 9545-8036      Fax: (+61) 3 9545-8080
!     e-mail: Alan.Miller @ mel.dms.csiro.au

!     The PUBLIC variables are:
!     lsq_kind = a KIND parameter for the floating-point quantities calculated
!                in this module.   See the more detailed explanation below.
!                This KIND parameter should be used for all floating-point
!                arguments passed to routines in this module.

!     nobs    = the number of observations processed to date.
!     ncol    = the total number of variables, including one for the constant,
!               if a constant is being fitted.
!     r_dim   = the dimension of array r = ncol*(ncol-1)/2
!     vorder  = an integer vector storing the current order of the variables
!               in the QR-factorization.   The initial order is 0, 1, 2, ...
!               if a constant is being fitted, or 1, 2, ... otherwise.
!     initialized = a logical variable which indicates whether space has
!                   been allocated for various arrays.
!     tol_set = a logical variable which is set when subroutine TOLSET has
!               been called to calculate tolerances for use in testing for
!               singularities.
!     rss_set = a logical variable indicating whether residual sums of squares
!               are available and usable.
!     d()     = array of row multipliers for the Cholesky factorization.
!               The factorization is X = Q.sqrt(D).R where Q is an ortho-
!               normal matrix which is NOT stored, D is a diagonal matrix
!               whose diagonal elements are stored in array d, and R is an
!               upper-triangular matrix with 1's as its diagonal elements.
!     rhs()   = vector of RHS projections (after scaling by sqrt(D)).
!               Thus Q'y = sqrt(D).rhs
!     r()     = the upper-triangular matrix R.   The upper triangle only,
!               excluding the implicit 1's on the diagonal, are stored by
!               rows.
!     tol()   = array of tolerances used in testing for singularities.
!     rss()   = array of residual sums of squares.   rss(i) is the residual
!               sum of squares with the first i variables in the model.
!               By changing the order of variables, the residual sums of
!               squares can be found for all possible subsets of the variables.
!               The residual sum of squares with NO variables in the model,
!               that is the total sum of squares of the y-values, can be
!               calculated as rss(1) + d(1)*rhs(1)^2.   If the first variable
!               is a constant, then rss(1) is the sum of squares of
!               (y - ybar) where ybar is the average value of y.
!     sserr   = residual sum of squares with all of the variables included.
!
!--------------------------------------------------------------------------

!     General declarations

use dal
use omallocation_mod

INTEGER, private             :: nobs, ncol, r_dim
INTEGER, private, ALLOCATABLE :: vorder(:)
LOGICAL, private             :: initialized = .false.,                  &
                                tol_set = .false., rss_set = .false.

!     Note. lsq_kind is being set to give at least 10 decimal digit
!           representation of floating point numbers.   This should be
!           adequate for most problems except the fitting of polynomials.
!           lsq_kind is being set so that the same code can be run on PCs
!           and Unix systems, which will usually represent floating-point
!           numbers in `double precision', and other systems with larger
!           word lengths which will give similar accuracy in `single
!           precision'.


REAL (double), private, ALLOCATABLE :: d(:), rhs(:), r(:), tol(:), rss(:)
REAL (double), private              :: zero1 = 0.0, one = 1.0,           &
                                vsmall

real(double) :: sserr
!PUBLIC                       :: nobs, ncol, r_dim, vorder,    &
!                                initialized, tol_set, rss_set,          &
!                                d, rhs, r, tol, rss, sserr
!PRIVATE                      :: zero1, one, vsmall

SAVE                         ! Save everything





!*************************************************************************
!
!
!*************************************************************************

CONTAINS


real(double) function getSserr()
    getSserr = sserr
end function getsserr

SUBROUTINE startup(nvar, fit_const)

!     Allocates dimensions for arrays and initializes to zero
!     The calling program must set nvar = the number of variables, and
!     fit_const = .true. if a constant is to be included in the model,
!     otherwise fit_const = .false.
!
!--------------------------------------------------------------------------

    IMPLICIT NONE
    INTEGER, INTENT(IN)            :: nvar
    LOGICAL, INTENT(IN)            :: fit_const

    !     Local variable
    integer :: i

    vsmall = 10. * TINY(zero1)

    nobs = 0
    IF (fit_const) THEN
        ncol = nvar + 1
    ELSE
        ncol = nvar
    END IF

    IF (initialized) DEALLOCATE(d, rhs, r, tol, rss, vorder)

    r_dim = ncol * (ncol - 1)/2

    ALLOCATE( d(ncol), rhs(ncol), r(r_dim), tol(ncol), rss(ncol), vorder(ncol) )

    d = zero1
    rhs = zero1
    r = zero1
    sserr = zero1

    IF (fit_const) THEN
        DO i = 1, ncol
            vorder(i) = i - 1
        END DO
    ELSE
        DO i = 1, ncol
            vorder(i) = i
       END DO
    END IF ! (fit_const)

    initialized = .true.
    tol_set = .false.
    rss_set = .false.

END SUBROUTINE startup
!****************************************************************************
!
!***************************************************************************
subroutine deleteLeastSquaresArrays


    if (initialized) then

        deallocate( d, rhs, r, tol, rss, vorder )

        initialized = .false.

    end if
end subroutine deleteLeastSquaresArrays

!****************************************************************************
!
!***************************************************************************


SUBROUTINE includ(weight, xrow, yelem)

!     ALGORITHM AS75.1  APPL. STATIST. (1974) VOL.23, NO. 3

!     Calling this routine updates D, R, RHS and SSERR by the
!     inclusion of xrow, yelem with the specified weight.

!     *** WARNING  Array XROW is overwritten.

!     N.B. As this routine will be called many times in most applications,
!          checks have been eliminated.
!
!--------------------------------------------------------------------------


    IMPLICIT NONE
    REAL (double), INTENT(IN)    :: weight, yelem
    REAL (double), INTENT(INOUT) :: xrow(ncol)

!     Local variables

    INTEGER           :: i, k, nextr
    REAL (double)   :: w, y, xi, di, wxi, dpi, cbar, sbar, xk

    nobs = nobs + 1
    w = weight
    y = yelem
    rss_set = .false.
    nextr = 1

    do i = 1, ncol
        !********************************************************************************
        !     Skip unnecessary transformations.   Test on exact zeroes must be
        !     used or stability can be destroyed.
        !******************************************************************************** 
        IF (ABS(w) < vsmall) RETURN
        xi = xrow(i)
        IF (ABS(xi) < vsmall) THEN
            nextr = nextr + ncol - i
        ELSE
            di = d(i)
            wxi = w * xi
            dpi = di + wxi * xi
            cbar = di / dpi
            sbar = wxi / dpi
            w = cbar * w
            d(i) = dpi
            DO k = i + 1, ncol
                xk = xrow(k)
                xrow(k) = xk - xi * r(nextr)
                r(nextr) = cbar * r(nextr) + sbar * xk
                nextr = nextr + 1
            END DO
            xk = y
            y = xk - xi * rhs(i)
           rhs(i) = cbar * rhs(i) + sbar * xk
        END IF
    END DO ! i = 1, ncol

!     Y * SQRT(W) is now equal to the Brown, Durbin & Evans recursive
!     residual.

    sserr = sserr + w * y * y


END SUBROUTINE includ
!************************************************************************
!
!
!************************************************************************
SUBROUTINE regcf(beta, nreq, ifault)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Modified version of AS75.4 to calculate regression coefficients
!     for the first NREQ variables, given an orthogonal reduction from
!     AS75.1.
!
!--------------------------------------------------------------------------

    IMPLICIT NONE
    INTEGER, INTENT(IN)            :: nreq
    INTEGER, INTENT(OUT)           :: ifault
    REAL (double) ,INTENT(OUT)   :: beta(nreq)

!     Local variables

INTEGER             :: i, j, nextr

!     Some checks.

ifault = 0
IF (nreq .LT. 1 .OR. nreq .GT. ncol) ifault = ifault + 4
IF (ifault .NE. 0) RETURN

IF (.NOT. tol_set) CALL tolset

DO i = nreq, 1, -1
  IF (SQRT(d(i)) .LT. tol(i)) THEN
    beta(i) = zero1
    d(i) = zero1
    ifault = -i
  ELSE
    beta(i) = rhs(i)
    nextr = (i-1) * (ncol+ncol-i)/2 + 1
    DO j = i+1, nreq
      beta(i) = beta(i) - r(nextr) * beta(j)
      nextr = nextr + 1
    END DO ! j = i+1, nreq
  END IF
END DO ! i = nreq, 1, -1

RETURN
END SUBROUTINE regcf
!***********************************************************************
!
!***********************************************************************
SUBROUTINE tolset

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Sets up array TOL for testing for zeroes in an orthogonal
!     reduction formed using AS75.1.

!     Local variables.
!
!--------------------------------------------------------------------------

IMPLICIT NONE

!     Local variables

INTEGER           :: col, row, pos
REAL (double)   :: eps, ten = 10.0, total, work(ncol)

!     EPS is a machine-dependent constant.

eps = ten * EPSILON(ten)

!     Set tol(i) = sum of absolute values in column I of R after
!     scaling each element by the square root of its row multiplier,
!     multiplied by EPS.

work = SQRT(d)
DO col = 1, ncol
  pos = col - 1
  total = work(col)
  DO row = 1, col-1
    total = total + ABS(r(pos)) * work(row)
    pos = pos + ncol - row - 1
  END DO
  tol(col) = eps * total
END DO

tol_set = .TRUE.
RETURN
END SUBROUTINE tolset
!********************************************************************
!
!
!********************************************************************
SUBROUTINE sing(lindep, ifault)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Checks for singularities, reports, and adjusts orthogonal
!     reductions produced by AS75.1.

!     Auxiliary routine called: INCLUD
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(OUT)        :: ifault
LOGICAL, INTENT(OUT)        :: lindep(ncol)

!     Local variables

REAL (double)   :: temp, x(ncol), work(ncol), y, weight
INTEGER           :: col, pos, row, pos2

ifault = 0

work = SQRT(d)

DO col = 1, ncol

!     Set elements within R to zero if they are less than tol(col) in
!     absolute value after being scaled by the square root of their row
!     multiplier.

  temp = tol(col)
  pos = col - 1
  DO row = 1, col-1
    IF (ABS(r(pos)) * work(row) .LT. temp) r(pos) = zero1
    pos = pos + ncol - row - 1
  END DO

!     If diagonal element is near zero, set it to zero, set appropriate
!     element of LINDEP, and use INCLUD to augment the projections in
!     the lower rows of the orthogonalization.

  lindep(col) = .FALSE.
  IF (work(col) .LE. temp) THEN
    lindep(col) = .TRUE.
    ifault = ifault - 1
    IF (col .LT. ncol) THEN
      pos2 = pos + ncol - col + 1
      x = zero1
      x(col+1:ncol) = r(pos+1:pos2-1)
      y = rhs(col)
      weight = d(col)
      r(pos+1:pos2-1) = zero1
      d(col) = zero1
      rhs(col) = zero1
      CALL includ(weight, x, y)
    ELSE
      sserr = sserr + d(col) * rhs(col)**2
    END IF ! (col .LT. ncol)
  END IF ! (work(col) .LE. temp)
END DO ! col = 1, ncol
RETURN
END SUBROUTINE sing
!*******************************************************************
!
!******************************************************************
SUBROUTINE ss

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Calculates partial residual sums of squares from an orthogonal
!     reduction from AS75.1.
!
!--------------------------------------------------------------------------

!     Local variables

IMPLICIT NONE
INTEGER           :: i
REAL (double)   :: total

total = sserr
rss(ncol) = sserr
DO i = ncol, 2, -1
  total = total + d(i) * rhs(i)**2
  rss(i-1) = total
END DO

rss_set = .TRUE.
RETURN
END SUBROUTINE ss
!*******************************************************************************************
!
!
!*******************************************************************************************



SUBROUTINE cov(nreq, var, covmat, dimcov, sterr, ifault)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Calculate covariance matrix for regression coefficients for the
!     first nreq variables, from an orthogonal reduction produced from
!     AS75.1.

!     Auxiliary routine called: INV
!
!--------------------------------------------------------------------------

    IMPLICIT NONE
    INTEGER, INTENT(IN)          :: nreq, dimcov
    INTEGER, INTENT(OUT)         :: ifault
    REAL (double), INTENT(OUT) :: var, covmat(dimcov), sterr(nreq)

    !     Local variables.

    INTEGER                      :: dim_rinv, pos, row, start, pos2, col, pos1, k
    REAL (double)              :: total
    REAL (double), ALLOCATABLE :: rinv(:)

    !     Check that dimension of array covmat is adequate.

    IF (dimcov < nreq * (nreq + 1) / 2) THEN
        ifault = 1
        RETURN
    END IF

    ! Check for small or zero multipliers on the diagonal.

    ifault = 0
    DO row = 1, nreq
        IF (ABS(d(row)) .LT. vsmall) ifault = -row
    END DO
    IF (ifault .NE. 0) RETURN

    ! Calculate estimate of the residual variance.

    IF (nobs > nreq) THEN
        var = rss(nreq) / (nobs - nreq)
    ELSE
       ifault = 2
        RETURN
    END IF

    dim_rinv = nreq * (nreq - 1) / 2
    ALLOCATE ( rinv(dim_rinv) )

    CALL INV(nreq, rinv)
    pos = 1
    start = 1

    DO row = 1, nreq
        pos2 = start
        DO col = row, nreq
            pos1 = start + col - row
            IF (row .EQ. col) THEN
                total = one / d(col)
            ELSE
                total = rinv(pos1-1) / d(col)
            END IF
            DO K = col + 1, nreq
                total = total + rinv(pos1) * rinv(pos2) / d(k)
                pos1 = pos1 + 1
                pos2 = pos2 + 1
           END DO ! K = col+1, nreq
           covmat(pos) = total * var
           IF (row .EQ. col) sterr(row) = SQRT(covmat(pos))
           pos = pos + 1
        END DO ! col = row, nreq
        start = start + nreq - row
    END DO ! row = 1, nreq


END SUBROUTINE cov
!*******************************************************************************************
!
!
!*******************************************************************************************



SUBROUTINE inv(nreq, rinv)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Invert first nreq rows and columns of Cholesky factorization
!     produced by AS 75.1.
!
!--------------------------------------------------------------------------

IMPLICIT NONE
    INTEGER, INTENT(IN)            :: nreq
    REAL (double), INTENT(OUT)   :: rinv(*)

    ! Local variables.

    INTEGER           :: pos, row, col, start, K, pos1, pos2
    REAL (double)   :: total

    ! Invert R ignoring row multipliers, from the bottom up.

    pos = nreq * (nreq - 1) / 2
    DO row = nreq - 1, 1, -1
        start = (row - 1) * (ncol + ncol - row) / 2 + 1
        DO col = nreq, row + 1, -1
            pos1 = start
            pos2 = pos
            total = zero1
            DO K = row + 1, col - 1
                pos2 = pos2 + nreq - K
                total = total - r(pos1) * rinv(pos2)
                pos1 = pos1 + 1
            END DO ! K = row+1, col-1
            rinv(pos) = total - r(pos1)
            pos = pos - 1
        END DO ! col = nreq, row+1, -1
    END DO ! row = nreq-1, 1, -1


END SUBROUTINE inv
!*******************************************************************************************
!
!
!*******************************************************************************************


SUBROUTINE partial_corr(in, cormat, dimc, ycorr, ifault)

!     Replaces subroutines PCORR and COR of:
!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Calculate partial correlations after the variables in rows
!     1, 2, ..., IN have been forced into the regression.
!     If IN = 1, and the first row of R represents a constant in the
!     model, then the usual simple correlations are returned.

!     If IN = 0, the value returned in array CORMAT for the correlation
!     of variables Xi & Xj is:
!       sum ( Xi.Xj ) / Sqrt ( sum (Xi^2) . sum (Xj^2) )

!     On return, array CORMAT contains the upper triangle of the matrix of
!     partial correlations stored by rows, excluding the 1's on the diagonal.
!     e.g. if IN = 2, the consecutive elements returned are:
!     (3,4) (3,5) ... (3,ncol), (4,5) (4,6) ... (4,ncol), etc.
!     Array YCORR stores the partial correlations with the Y-variable
!     starting with YCORR(IN+1) = partial correlation with the variable in
!     position (IN+1).
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)          :: in, dimc
INTEGER, INTENT(OUT)         :: ifault
REAL (double), INTENT(OUT) :: cormat(dimc), ycorr(ncol)

!     Local variables.

INTEGER           :: base_pos, pos, row, col, col1, col2, pos1, pos2
REAL (double)   :: rms(in+1:ncol), sumxx, sumxy, sumyy, work(in+1:ncol)

!     Some checks.

ifault = 0
IF (in .LT. 0 .OR. in .GT. ncol-1) ifault = ifault + 4
IF (dimc .LT. (ncol-in)*(ncol-in-1)/2) ifault = ifault + 8
IF (ifault .NE. 0) RETURN

!     Base position for calculating positions of elements in row (IN+1) of R.

base_pos = in*ncol - (in+1)*(in+2)/2

!     Calculate 1/RMS of elements in columns from IN to (ncol-1).

IF (d(in+1) > zero1) rms(in+1) = one / SQRT(d(in+1))
DO col = in+2, ncol
  pos = base_pos + col
  sumxx = d(col)
  DO row = in+1, col-1
    sumxx = sumxx + d(row) * r(pos)**2
    pos = pos + ncol - row - 1
  END DO ! row = in+1, col-1
  IF (sumxx > zero1) THEN
    rms(col) = one / SQRT(sumxx)
  ELSE
    rms(col) = zero1
    ifault = -col
  END IF ! (sumxx > zero)
END DO ! col = in+1, ncol-1

!     Calculate 1/RMS for the Y-variable

sumyy = sserr
DO row = in+1, ncol
  sumyy = sumyy + d(row) * rhs(row)**2
END DO ! row = in+1, ncol
IF (sumyy > zero1) sumyy = one / SQRT(sumyy)

!     Calculate sums of cross-products.
!     These are obtained by taking dot products of pairs of columns of R,
!     but with the product for each row multiplied by the row multiplier
!     in array D.

pos = 1
DO col1 = in+1, ncol
  sumxy = zero1
  work(col1+1:ncol) = zero1
  pos1 = base_pos + col1
  DO row = in+1, col1-1
    pos2 = pos1 + 1
    DO col2 = col1+1, ncol
      work(col2) = work(col2) + d(row) * r(pos1) * r(pos2)
      pos2 = pos2 + 1
    END DO ! col2 = col1+1, ncol
    sumxy = sumxy + d(row) * r(pos1) * rhs(row)
    pos1 = pos1 + ncol - row - 1
  END DO ! row = in+1, col1-1

!     Row COL1 has an implicit 1 as its first element (in column COL1)

  pos2 = pos1 + 1
  DO col2 = col1+1, ncol
    work(col2) = work(col2) + d(col1) * r(pos2)
    pos2 = pos2 + 1
    cormat(pos) = work(col2) * rms(col1) * rms(col2)
    pos = pos + 1
  END DO ! col2 = col1+1, ncol
  sumxy = sumxy + d(col1) * rhs(col1)
  ycorr(col1) = sumxy * rms(col1) * sumyy
END DO ! col1 = in+1, ncol-1

ycorr(1:in) = zero1


END SUBROUTINE partial_corr
!****************************************************************************
!
!
!****************************************************************************
SUBROUTINE vmove(from, to, ifault)

!     ALGORITHM AS274 APPL. STATIST. (1992) VOL.41, NO. 2

!     Move variable from position FROM to position TO in an
!     orthogonal reduction produced by AS75.1.
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)            :: from, to
INTEGER, INTENT(OUT)           :: ifault

!     Local variables

REAL (double)   :: d1, d2, X, d1new, d2new, cbar, sbar, Y
INTEGER           :: m, first, last, inc, m1, m2, mp1, col, pos, row

!     Check input parameters

ifault = 0
IF (from .LT. 1 .OR. from .GT. ncol) ifault = ifault + 4
IF (to .LT. 1 .OR. to .GT. ncol) ifault = ifault + 8
IF (ifault .NE. 0) RETURN

IF (from .EQ. to) RETURN

IF (.NOT. rss_set) CALL ss

IF (from .LT. to) THEN
  first = from
  last = to - 1
  inc = 1
ELSE
  first = from - 1
  last = to
  inc = -1
END IF

   DO 70 m = first, last, inc

!     Find addresses of first elements of R in rows M and (M+1).

     m1 = (m-1)*(ncol+ncol-m)/2 + 1
     m2 = m1 + ncol - m
     mp1 = m + 1
     d1 = d(m)
     d2 = d(mp1)

!     Special cases.

     IF (d1 .LT. vsmall .AND. d2 .LT. vsmall) GO TO 40
     X = r(m1)
     IF (ABS(x) * SQRT(d1) .LT. tol(mp1)) THEN
       X = zero1
     END IF
     IF (d1 .LT. vsmall .OR. ABS(x) .LT. vsmall) THEN
       d(m) = d2
       d(mp1) = d1
       r(m1) = zero1
       DO col = m+2, ncol
         m1 = m1 + 1
         X = r(m1)
         r(m1) = r(m2)
         r(m2) = x
         m2 = m2 + 1
       END DO ! col = m+2, ncol
       X = rhs(m)
       rhs(m) = rhs(mp1)
       rhs(mp1) = x
       GO TO 40
     ELSE IF (d2 .LT. vsmall) THEN
       d(m) = d1 * x**2
       r(m1) = one / x
       r(m1+1:m1+ncol-m-1) = r(m1+1:m1+ncol-m-1) / x
       rhs(m) = rhs(m) / x
       GO TO 40
     END IF
!
!     Planar rotation in regular case.
!
     d1new = d2 + d1*x**2
     cbar = d2 / d1new
     sbar = x * d1 / d1new
     d2new = d1 * cbar
     d(m) = d1new
     d(mp1) = d2new
     r(m1) = sbar
     DO col = m+2, ncol
       m1 = m1 + 1
       y = r(m1)
       r(m1) = cbar*r(m2) + sbar*y
       r(m2) = y - x*r(m2)
       m2 = m2 + 1
     END DO ! col = m+2, ncol
     y = rhs(m)
     rhs(m) = cbar*rhs(mp1) + sbar*y
     rhs(mp1) = y - x*rhs(mp1)

!     Swap columns M and (M+1) down to row (M-1).

40   IF (m .EQ. 1) GO TO 60
     pos = m
     DO row = 1, m-1
       x = r(pos)
       r(pos) = r(pos-1)
       r(pos-1) = x
       pos = pos + ncol - row - 1
     END DO ! row = 1, m-1

!     Adjust variable order (VORDER), the tolerances (TOL) and
!     the vector of residual sums of squares (RSS).

60   m1 = vorder(m)
     vorder(m) = vorder(mp1)
     vorder(mp1) = m1
     x = tol(m)
     tol(m) = tol(mp1)
     tol(mp1) = x
     rss(m) = rss(mp1) + d(mp1) * rhs(mp1)**2
70 CONTINUE


END SUBROUTINE vmove
!**************************************************************************
!
!
!*************************************************************************
SUBROUTINE reordr(list, n, pos1, ifault)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2

!     Re-order the variables in an orthogonal reduction produced by
!     AS75.1 so that the N variables in LIST start at position POS1,
!     though will not necessarily be in the same order as in LIST.
!     Any variables in VORDER before position POS1 are not moved.

!     Auxiliary routine called: VMOVE
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)         :: n, list(n), pos1
INTEGER, INTENT(OUT)        :: ifault

!     Local variables.

INTEGER        :: next, i, l, j

!     Check N.

   ifault = 0
   IF (n .LT. 1 .OR. n .GT. ncol+1-pos1) ifault = ifault + 4
   IF (ifault .NE. 0) RETURN

!     Work through VORDER finding variables which are in LIST.

   next = pos1
   i = pos1
10 l = vorder(i)
   DO j = 1, n
     IF (l .EQ. list(j)) GO TO 40
   END DO
30 i = i + 1
   IF (i .LE. ncol) GO TO 10

!     If this point is reached, one or more variables in LIST has not
!     been found.

   ifault = 8
   RETURN

!     Variable L is in LIST; move it up to position NEXT if it is not
!     already there.

40 IF (i .GT. next) CALL vmove(i, next, ifault)
   next = next + 1
   IF (next .LT. n+pos1) GO TO 30

   
END SUBROUTINE reordr
!*****************************************************************************
!
!
!****************************************************************************
SUBROUTINE hdiag(xrow, nreq, hii, ifault)

!     ALGORITHM AS274  APPL. STATIST. (1992) VOL.41, NO. 2
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)          :: nreq
INTEGER, INTENT(OUT)         :: ifault
REAL (double), INTENT(IN)  :: xrow(ncol)
REAL (double), INTENT(OUT) :: hii

!     Local variables

INTEGER             :: col, row, pos
REAL (double)     :: total, wk(ncol)

!     Some checks

ifault = 0
IF (nreq .GT. ncol) ifault = ifault + 4
IF (ifault .NE. 0) RETURN

!     The elements of xrow.inv(R).sqrt(D) are calculated and stored
!     in WK.

hii = zero1
DO col = 1, nreq
  IF (SQRT(d(col)) .LE. tol(col)) THEN
    wk(col) = zero1
  ELSE
    pos = col - 1
    total = xrow(col)
    DO row = 1, col-1
      total = total - wk(row)*r(pos)
      pos = pos + ncol - row - 1
    END DO ! row = 1, col-1
    wk(col) = total
    hii = hii + total**2 / d(col)
  END IF
END DO ! col = 1, nreq


END SUBROUTINE hdiag
!***********************************************************************
!
!
!***********************************************************************
REAL (double) FUNCTION varprd(x, nreq, var, ifault)

!     Calculate the variance of x'b where b consists of the first nreq
!     least-squares regression coefficients.
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)          :: nreq
INTEGER, INTENT(OUT)         :: ifault
REAL (double), INTENT(IN)  :: x(*)
REAL (double), INTENT(OUT) :: var

!     Local variables

INTEGER           :: row
REAL (double)   :: wk(nreq)

!     Check input parameter values

varprd = zero1
ifault = 0
IF (nreq .LT. 1 .OR. nreq .GT. ncol) ifault = ifault + 4
IF (nobs .LE. nreq) ifault = ifault + 8
IF (ifault .NE. 0) RETURN

!     Calculate the residual variance estimate.

var = sserr / (nobs - nreq)

!     Variance of x'b = var.x'(inv R)(inv D)(inv R')x
!     First call BKSUB2 to calculate (inv R')x by back-substitution.

CALL BKSUB2(x, wk, nreq)
DO row = 1, nreq
  IF(d(row) .GT. tol(row)) varprd = varprd + wk(row)**2 / d(row)
END DO

varprd = varprd * var

RETURN
END FUNCTION varprd
!*********************************************************************************
!
!
!*********************************************************************************


SUBROUTINE bksub2(x, b, nreq)

!     Solve x = R'b for b given x, using only the first nreq rows and
!     columns of R, and only the first nreq elements of R.
!
!--------------------------------------------------------------------------

IMPLICIT NONE
INTEGER, INTENT(IN)          :: nreq
REAL (double), INTENT(IN)  :: x(nreq)
REAL (double), INTENT(OUT) :: b(nreq)

!     Local variables

INTEGER           :: pos, row, col
REAL (double)   :: temp

!     Solve by back-substitution, starting from the top.

DO row = 1, nreq
  pos = row - 1
  temp = x(row)
  DO col = 1, row-1
    temp = temp - r(pos)*b(col)
    pos = pos + ncol - col - 1
  END DO
  b(row) = temp
END DO


END SUBROUTINE bksub2

!****************************************************************************
!
!****************************************************************************

SUBROUTINE separate_text(text, name, number)

!     Takes the character string in `text' and separates it into `names'.
!     This version only allows spaces as delimiters.

IMPLICIT NONE
CHARACTER (LEN = *), INTENT(INOUT)  :: text
CHARACTER (LEN = *), INTENT(OUT)    :: name(*)
INTEGER, INTENT(OUT)                :: number

!     Local variables

INTEGER  :: length, len_name, pos1, pos2, nchars

length = LEN_TRIM(text)
number = 0
IF (length == 0) RETURN

len_name = LEN(name(1))
IF (len_name < 1) RETURN

pos1 = 1
DO
  DO                                   ! Remove any leading blanks
    IF (text(pos1:pos1) == ' ') THEN
      text(pos1:) = text(pos1+1:)
      length = length - 1
      CYCLE
    ELSE
      EXIT
    END IF
  END DO
  pos2 = INDEX(text(pos1:length), ' ') ! Find position of next blank
  IF (pos2 > 0) THEN
    pos2 = pos2 + pos1 - 1
  ELSE                                 ! Last name, no blank found
    pos2 = length + 1
  END IF
  number = number + 1
  nchars = MIN(len_name, pos2-pos1)
  name(number) = text(pos1:pos1+nchars-1)
  pos1 = pos2 + 1                      ! Move past the blank
  IF (pos1 > length) RETURN
END DO

END SUBROUTINE separate_text
!*******************************************************************
!
!
!*******************************************************************


SUBROUTINE printc(in, cormat, dimc, ycorr, vname, yname, iopt, lout, ier)

!     Print (partial) correlations calculated using partial_corr to unit lout.
!     If IOPT = 0, print correlations with the Y-variable only.



IMPLICIT NONE
INTEGER, INTENT(IN)           :: in, dimc, iopt, lout
INTEGER, INTENT(OUT)          :: ier
REAL ( double ), INTENT(IN) :: cormat(dimc), ycorr(ncol)
CHARACTER, INTENT(IN)         :: vname(0:ncol-1)*8, yname*8

!     Local variables.

INTEGER            :: nrows, j1, j2, i1, i2, row, upos, tpos, last
CHARACTER          :: text*74, empty*65 = ' ', char1*9 = ' 1.0'

!     Check validity of arguments

ier = 0
IF (in .GE. ncol) ier = 1
IF (ncol .LE. 1) ier = ier + 2
nrows = ncol - in
IF (dimc .LE. nrows*(nrows-1)/2) ier = ier + 4
IF (ier .NE. 0) RETURN

!     If iopt.NE.0 output heading

IF (iopt .EQ. 0) GO TO 30
WRITE(lout, 900)
900 FORMAT(/5x, 'Correlation matrix')
j1 = in + 1

DO
  j2 = MIN(j1+6, ncol)
  i1 = j1 - in
  i2 = j2 - in
  WRITE(lout, 910) vname(vorder(j1:j2))
  910 FORMAT(11X, 7(a8, 1x))

!     Print correlations for rows 1 to i2, columns i1 to i2.

  DO row = 1, i2
    text = ' ' // vname(vorder(row+in)) // empty
    IF (i1 .GT. row) THEN
      upos = (row-1) * (nrows+nrows-row) /2 + (i1-row)
      last = upos + i2 - i1
      WRITE(text(12:74), '(7(F8.5, 1x))') cormat(upos:last)
    ELSE
      upos = (row-1) * (nrows+nrows-row) /2 + 1
      tpos = 12 + 9*(row-i1)
      text(tpos:tpos+8) = char1
      last = upos + i2 - row - 1
      IF (row .LT. i2) WRITE(text(tpos+9:74), '(6(F8.5, 1X))') cormat(upos:last)
    END IF
    WRITE(lout, '(a)') text
  END DO ! row = 1, i2

!     Move onto the next block of columns.

  j1 = j2 + 1
  IF (j1 .GT. ncol) EXIT
END DO

!     Correlations with the Y-variable.

30 WRITE(lout, 920) yname
   920 FORMAT(/5x, 'Correlations with the dependent variable: ', a)
   j1 = in + 1
DO
  j2 = MIN(j1+7, ncol)
  WRITE(lout, 930) vname(vorder(j1:j2))
  930 FORMAT(/1X, 8(A8, 1X))
  WRITE(lout, 940) ycorr(j1:j2)
  940 FORMAT(1X, 8(F8.5, 1X))
  j1 = j2 + 1
  IF (j1 .GT. ncol) EXIT
END DO

END SUBROUTINE printc

!*******************************************************************
!
!
!******************************************************************
subroutine LeastSquaresFit(n, xData, ydata, weight, maxIter, gradient, &
           gradientError, constant, constantError)
    integer, intent(in) :: n
    real(single) :: xData(:), ydata(:), weight(:)
    integer(int8), intent(in), optional :: maxIter
    real(single), intent(out) :: gradient, gradientError, constant, constantError
    integer :: i, j, n1, m, ifail, nCoeffs, nits
  
    real(double) :: beta(1 : 4), x(1 : 4), var(1 : 20)
    real(double) :: d, sterr(5)
!    character(200) :: text    
    real(single), pointer :: residual(:) => null()
    real(single) :: std
    
    std = 1000000.0



    residual => allocateSingle(1, n, "residual")


    
    residual = 0.0
    std = 100000.0

    if(present(maxIter)) then
        nits = maxIter
    else     
        nits = 2
    end if
    n1 = n

    do j = 1, nits
        nCoeffs = 2
        call startup(nCoeffs, .false.)
        call tolset
        m = 0

        do i = 1, n
            if (residual(i) .gt. 2.0 * std) cycle
            
            m = m + 1
            x(1) = 1.0_double
            x(2) = dble(xdata(i))
          
  
            call includ(dble(weight(i)), x, dble(ydata(i)))


        end do
        if (j .gt. 1 .and. (m .lt. n / 2 .or. m  .eq. n1)) exit
        n1 = m  
        call regcf(beta, 2, ifail)
        call cov(2, d, var, 20, sterr, ifail)
   
        var = sqrt(var)
!        write(*, *) "Computed regression coefficients- iteration ", j
!        write(text, '(a, f10.4, a, f10.4, a, f10.4, a, f10.4, a, a, f10.4, a, i6)')  "Regression line - y=(", beta(1), &
!                                         "+/-", var(1), ") + (", beta(2), "+/-", var(2), ")*x", &
!                 " residual sum of squares=", sserr, " n=", n1

       
!        write(*, *) text
        do i = 1, n
            residual(i) = (yData(i) - (beta(1) + beta(2) * xdata(i))) * weight(i)
        end do
        std = sqrt(sserr / real(m - 1))
    end do 

    constant = beta(1)
    constantError = sterr(1)

    gradient = beta(2)
    gradientError = sterr(2)

    deallocate(residual)
    nullify(residual)


end subroutine LeastSquaresFit




!*******************************************
!
!*****************************************
subroutine LeastSquaresFit1(m, modelData, fitData, weight, gradient, error, residual1)
    integer, intent(in) :: m
    real(single), dimension(:) :: modelData, fitData, weight
    real(single), intent(out) :: gradient, error
    real(single), optional, dimension(:) :: residual1 
    real(single), pointer :: residual(:) => null()

    integer :: i
    real(double) :: sum1, sum2 
!    character(200) :: text  
    sum1 = 0.0_double
    sum2 = 0.0_double
    
    do i = 1, m
        sum1 = sum1 + weight(i) * dble(modelData(i)) * dble(fitdata(i))
        sum2 = sum2 + weight(i) * dble(modelData(i)) * dble(modelData(i))
      
    
    end do

    gradient = sum1 / sum2  

    ! Compute the error

  
    ! Error in fit constant
    ! Formula used is that from P. Van de kamp, 1967, Principles of Astrometry, W.H. Freeman and Company, page. 217.

    residual => allocateSingle(1, m, "residual")
    sum1 = 0.0_double
    sum2 = 0.0_double

    do i = 1, m
        residual(i) = (fitdata(i) - modeldata(i) * gradient) * weight(i)
        sum1 = sum1 + weight(i) * residual(i) * residual(i)
        sum2 = sum2 + weight(i) * modeldata(i) * modeldata(i)
               
    end do

    error = 0.6745 * sqrt(sum1 / dble(m - 1)) / sqrt(sum2) ! The probable error


!    write(*, *) "Computed regression coefficients- iteration "

!    write(text, '(a, f10.4, a, f10.4, a, a, f10.4, a, i6)')  &
!                  "Regression line - y=(", gradient, "+/-", error, ")*x", &
!                 " residual sum of squares=", sum1, " n=", m

!    write(*, *) trim(text)




    if(present(residual1)) residual1(1 : m) = residual(1 : m)
    deallocate(residual)



end subroutine leastsquaresfit1



!*******************************************************************
! Least-squares fit of the kind y=constant+a1*x+a2*x etc
! nterms is the number of a constants to be solved for
!
!******************************************************************
subroutine LeastSquaresFit2(nData, xData, ydata, weight, nterms, maxIter, &
            nstd1, gradient, gradientError, constant, constantError)
    integer, intent(in) :: nData
    real(single) :: xData(:, :), ydata(:), weight(:)
    integer(int8), intent(in) :: nTerms
    integer(int8), intent(in), optional :: maxIter
    real(single), intent(in), optional :: nstd1
    real(single), intent(out), dimension(:) :: gradient, gradientError
    real(single), intent(out) :: constant, constantError
    integer :: i, j, k, n1, m, ifail, nCoeffs, nits
  
    real(double), dimension(nterms + 1) :: beta, x
    real(double), dimension(:), allocatable :: sterr, var    
    real(single), pointer :: residual(:) => null()
    real(single) :: std, nstd
    real(double) :: fit, d    




    if(present(nstd1)) then
        nstd = nstd1
    else
        nstd = 3.0
    end if
    std = 1000000.0

    residual => allocateSingle(1, nData, "residual")
    
    residual = 0.0
    std = 100000.0
    if(present(maxIter)) then
        nits = maxIter
    else     
        nits = 2
    end if
    n1 = nData
    nCoeffs = nterms + 1
 
    m = ncoeffs * (ncoeffs + 1)
    allocate(sterr(1 : m))  

    allocate(var(1 : m))  


   
    do j = 1, nits
        
        call startup(nCoeffs, .false.)
        call tolset  


        m = 0

        do i = 1, nData
            if (residual(i) .gt. nstd * std) cycle
            
            m = m + 1
            
            x(1) = 1.0_double
            do k = 2, nCoeffs
                x(k) = dble(xdata(i, k - 1))
            end do
          
            call includ(dble(weight(i)), x, dble(ydata(i)))


        end do
        if (j .gt. 1 .and. (m .lt. nData / 2 .or. m  .eq. n1)) exit
        n1 = m  


        !*****************************************************************
        ! Compute the regression coefficients
        !*****************************************************************
        call regcf(beta, ncoeffs, ifail)


        call cov(ncoeffs, d, var, size(var), sterr, ifail)
!        if(ifail .ne. 0) then
!            write(*, *) "Error! ifail=", ifail
!        endif   

        var = sqrt(var)

        do i = 1, nData
            fit = beta(1)
            do k = 2, nCoeffs
                fit = fit + dble(xdata(i, k - 1)) * beta(k)
            end do  
            residual(i) = (yData(i) - fit) * weight(i)
        end do
         
        std = sqrt(sserr / real(m - 1))
    end do 


    constant = beta(1)
    constantError = sterr(1)
!    write(*, *) "Background=", constant, "+/-", constantError
    do k = 1, nterms
        gradient(k) = beta(k + 1)
        gradientError(k) = sterr(k + 1)
!        write(*, *) "Least-squares fit 2 - gradient ", k, ")=", gradient(k), "+/-", gradientError(k)
    end do


 
    deallocate(residual)
    nullify(residual)
    deallocate(sterr, var)

end subroutine LeastSquaresFit2




!*******************************************************************
! Least-squares fit of the kind y=a1*x+a2*x etc
! nterms is the number of a constants to be solved for
!
!******************************************************************
subroutine LeastSquaresFit3(nData, xData, ydata, weight, nterms, maxIter, &
            nstd1, gradient, gradientError)
    integer, intent(in) :: nData
    real(single) :: xData(:, :), ydata(:), weight(:)
    integer(int8), intent(in) :: nTerms
    integer(int8), intent(in), optional :: maxIter
    real(single), intent(in), optional :: nstd1
    real(single), intent(out), dimension(:) :: gradient, gradientError
   
    integer :: i, j, k, n1, m, ifail, nCoeffs, nits
  
    real(double), dimension(nterms + 1) :: beta, x
    real(double), dimension(:), allocatable :: sterr, var    
    real(single), pointer :: residual(:) => null()
    real(single) :: std, nstd
    real(double) :: fit, d    




    if(present(nstd1)) then
        nstd = nstd1
    else
        nstd = 3.0
    end if
    std = 1000000.0

    residual => allocateSingle(1, nData, "residual")
    
    residual = 0.0
    std = 100000.0
    if(present(maxIter)) then
        nits = maxIter
    else     
        nits = 2
    end if
    n1 = nData
    nCoeffs = nterms
 
    m = ncoeffs * (ncoeffs + 1)
    allocate(sterr(1 : m))  

    allocate(var(1 : m))  


   
    do j = 1, nits
        
        call startup(nCoeffs, .false.)
        call tolset  


        m = 0

        do i = 1, nData
            if (residual(i) .gt. nstd * std) cycle
            
            m = m + 1
            
            
            do k = 1, nCoeffs
                x(k) = dble(xdata(i, k))
            end do
          
            call includ(dble(weight(i)), x, dble(ydata(i)))


        end do
        if (j .gt. 1 .and. (m .lt. nData / 2 .or. m  .eq. n1)) exit
        n1 = m  


        !*****************************************************************
        ! Compute the regression coefficients
        !*****************************************************************
        call regcf(beta, ncoeffs, ifail)


        call cov(ncoeffs, d, var, size(var), sterr, ifail)
!        if(ifail .ne. 0) then
!            write(*, *) "Error! ifail=", ifail
!        endif   

        var = sqrt(var)

        do i = 1, nData
            fit = beta(1)
            do k = 2, nCoeffs
                fit = fit + dble(xdata(i, k - 1)) * beta(k)
            end do  
            residual(i) = (yData(i) - fit) * weight(i)
        end do
         
        std = sqrt(sserr / real(m - 1))
    end do 


  
!    write(*, *) "Background=", constant, "+/-", constantError
    do k = 1, nterms
        gradient(k) = beta(k)
        gradientError(k) = sterr(k)
!        write(*, *) "Least-squares fit 2 - gradient ", k, ")=", gradient(k), "+/-", gradientError(k)
    end do


 
    deallocate(residual)
    nullify(residual)
    deallocate(sterr, var)

end subroutine LeastSquaresFit3







!*******************************************************************
!
!
!******************************************************************
subroutine LeastSquaresPolyFit(ndata, xData, ydata, dyData, order, nits1, &
            xCoeff, dxCoeff, std)
    integer, intent(in) :: ndata
    real(single) :: xData(:), ydata(:), dyData(:)
    real(single), intent(out) :: std
    integer(int8), intent(in) :: order
    integer(int8), optional :: nits1
    real(single) :: xCoeff(:), dxCoeff(:)
    integer :: i, j, k, l, n1, m, ifail, nCoeffs, nits
    real(double) :: weight
    real(double) :: beta(1 : 20), x(1 : 20), var(1 : 20)
    real(double) :: d, sterr(20)
 
    real(single), pointer :: residual(:) => null()
   
   
    std = 1000.0

    if (present(nits1)) then
        nits = nits1
    else
        nits = 1
    end if
   
    
    residual => allocateSingle(1, ndata, "residual")
    
    residual = 0.0
    std = 100000.0
   
  
    n1 = ndata

    do j = 1, nits
        nCoeffs = order + 1
        call startup(nCoeffs, .true.)  ! true specifies using a constant
        call tolset
        m = 0

        do i = 1, ndata
            if (residual(i) .gt. 2.0 * std) cycle
            if (dyData(i) .gt. 0.0) then
                weight = 1.0_double / (dyData(i) * dyData(i))
            else
                weight = 1.0_double
            end if

            m = m + 1
            
            x(1) = 1.0_double
            l = 2
            x(2) = xdata(i)
            do k = 3, ncoeffs
                l = l + 1
                x(l) = x(l - 1) * xdata(i) 
            end do 
            call includ(weight, x, dble(ydata(i)))


        end do
        if (j .gt. 1 .and. (m .lt. ndata / 2 .or. m  .eq. n1)) exit
        n1 = m  
        call regcf(beta, ncoeffs, ifail)
        call cov(ncoeffs, d, var, 20, sterr, ifail)
   
        var = sqrt(var)


!        write(*, *) "Computed regression coefficients- iteration ", j
        do k = 1, ncoeffs
!            write(*, *) k, ") x=", beta(k), " +/-", sterr(k)
            xcoeff(k) = beta(k)
            dxcoeff(k) = sterr(k)

        end do

        
        
      
        do i = 1, ndata
            residual(i) = yData(i) - (beta(1) + beta(2) * xdata(i))
        end do
        std = sqrt(sserr / real(m - 1))





    end do 

 
    call deleteLeastSquaresArrays
    deallocate(residual)
    nullify(residual)


end subroutine LeastSquaresPolyFit

!*********************************************************
!
!
!
!********************************************************
real(single) Function EvaluatePolyNomial(x, order, coeff)
    real(single), intent(in) :: x
    integer(int8), intent(in) :: order
    real(single) :: coeff(:)
    real(double) :: sum, f
    integer(int8) :: i

 
    f = dble(x) 
 
    sum = dble(coeff(2)) * f + dble(coeff(1))


    do i = 3, order + 1_int8
        f = f * dble(x)
        sum = sum + dble(coeff(i)) * f
   
    end do
    
    EvaluatePolyNomial = real(sum)    
    

end function EvaluatePolyNomial

end module leastsquares_mod








