! !
! Authors:  Christopher Brindle - MSSL
!           Andrew King - MSSL
!           Tim Harries - MSSL
!          
! Description: Combines OSW source list into an observation source list
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
!   
!
!************************************************************************************************
module omsourceflags_mod

  use cal
  use dal
  use Errorhandling
  use omallocation_mod
  use omconstants_mod
  use omutility_mod
  use ominterpolate_mod
  use omnumerics_mod
  use omellipse_mod
  implicit none 


public ::  LookForReadOutStreaks
public ::  LookForStarSpikes
public ::  setSourceFlags 

interface ConvertCoordsToFullFrame
    module procedure ConvertCoordsToFullFrame1,  &
                     ConvertCoordsToFullFrame2, &
                     ConvertCoordsToFullFrame3

end interface
interface ReadOutStreaks
    module procedure ReadOutStreaks1, ReadOutStreaks2

end interface

interface SmokeRings
    module procedure SmokeRings1, SmokeRings2

end interface
    
  
interface BrightSourceProximityFlags
    module procedure BrightSourceProximityFlags1, BrightSourceProximityFlags2 

end interface



!****************************************************************************
! QFLAG source quality flags convention
! BIT NO             FLAGGING                         Integer
!
!   0                BAD PIXEL                                 1
!   1                READ-OUT STREAK                           2
!   2                SMOKE-RING                                4
!   3                SOURCE ON STAR-SPIKE                      8
!   4                Mod-8 source                             16
!   5                Source within central enhancement        32
!   6                Source lies near to a bright source      64
!   7                Near an edge                            128 
!   8                point-source within extended source     256 
!   9                Weird source                            512


!*****************************************************************************

!****************************************************************************
! CFLAG source confusion flags convention
! BIT NO             FLAGGING                                                Integer
!
!   0                one or more source within 0.5-1.0 aperture radius       1 
!   1                one or more source within 0.5 aperture radius           2
!*****************************************************************************

integer(int16), private, parameter :: noQFlags = 10
character(110), private, dimension(noQflags) :: qFlagsText =                                                    & 
(/ "0) Source on a bad pixel                       (red)                                                   ",   &
   "1) Source on a read-out streak                 (magenta)                                               ",   &
   "2) Source near or within a smoke-ring          (yellow)                                                ",   &
   "3) Source on or near a diffraction-spike       (white)                                                 ",   &
   "4) Source surrounded by a strong mod-8 pattern (black)                                                 ",   & 
   "5) Source within central enhanced region       (red)                                                   ",   &
   "6) Source lies near to a bright source         (black)                                                 ",   &
   "7) Point source lies with 1.0 aperture radius of edge/some point of extended source lies on edge (blue)",   &
   "8) Point-source lies partly or entirely within extended source region (blue)                           ",   &
   "9) Compact source/exceptionally strange        (red)                                                   "/)    




!*******************************************************************************************
! CONFUSION FLAGS
!*******************************************************************************************
integer(int8), private, parameter :: noCFlags = 2


character(110), private, dimension(noCFlags) :: cFlagsText = (/              & 
"0) One or more source within 0.5-1.0 aperture radius                     ", &   
"1) one or more source within 0.5 aperture radius                         "/)
                                                                                                



contains
!**********************************************************************************
!
!*********************************************************************************
integer(int16) function GetNoQFlags()
    GetNoQFlags = noQflags

end function GetNoQFlags

!**********************************************************************************
!
!*********************************************************************************
integer(int8) function GetNoCFlags()
    GetNoCFlags = noCflags

end function GetNoCFlags




!**********************************************************************************
!
!*********************************************************************************
function getQFlagText( i ) result(text)
    integer, intent(in) :: i
    character( len(trim(qflagsText(i)))) :: text
    text = trim(qflagsText(i) )
end function getQFlagText

!**********************************************************************************
!
!*********************************************************************************
function getCFlagText( i ) result(text)
    integer, intent(in) :: i
    character( len(trim(cflagsText(i)))) :: text
    text = trim(cflagsText(i) )
end function getCFlagText




!****************************************************
!
!***************************************************
subroutine SetSourceFlags(n, source, x0, y0, plevelimage, elevelimage)
    integer(int16), intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: x0, y0
   
 
    integer(int16), optional, pointer, dimension(:, :) ::  plevelimage, elevelimage
  
    real(single), pointer :: xCoord(:) => null(), yCoord(:) => null()
    integer :: i
    integer(int16) :: nx, ny
    if (.not. associated(source) .or. n .lt. 1) return

    
    xCoord => allocateSingle(1_int16, n, "xCoord")
    yCoord => allocateSingle(1_int16, n, "yCoord")

    xCoord(1 : n) = source(1 : n)%xPos
    yCoord(1 : n) = source(1 : n)%yPos

    nx = ubound(elevelimage, 1)
    ny = ubound(elevelimage, 2)
  
    call ConvertCoordsToFullFrame(xCoord, yCoord, x0, y0, binx, biny)





    call setCornerFlags(xCoord, yCoord, source, n)



    call setCentralRegionFlags(xCoord, ycoord, source, n)


    call starspikes(int(n, int32), source, xcoord, ycoord)



    call ReadOutStreaks(int(n, int32),          &
                  source%rawCountRate,          &
                  xCoord,                       &
                  source%extensionFlag,         &
                  source%qualityFlag)

  
    !***************************************************************
    ! Look for sources within the smoke-rings found by omdetect
    !
    !**************************************************************
    call setSmokeRings( n, source)
  


    call SmokeRings(int(n, int32),              &
                  source%rawCountRate,          &
                  xCoord, yCoord,               &
                  source%qualityFlag)




!    call LookForReadOutStreaks(source, image)


!    call LookForStarSpikes(source, image, 20 / binx)




    !**************************************************
    ! Flag sources close to a bright source
    !*************************************************
    call BrightSourceProximityFlags(int(n, int32),      &
                  source%corrCountRate,                 &
                  xCoord, yCoord,                       &
                  source%axialRatio,                    &
                  source%qualityFlag)


 
    deallocate(xCoord, ycoord)
    nullify(xCoord, ycoord)

    if (present(plevelimage) .and. present(elevelImage)) then 
      
        do i = 1, n
  
 
            if (source(i)%extensionFlag .eq. 0 .and. .not. btest(source(i)%qualityFlag, 4) ) then
                ! Check for any point sources within an extended source region

                if( source(i)%imin .ge. 0 .and. source(i)%jmin .ge. 0 .and. &
                  source(i)%imax .le. nx .and. source(i)%jmax .le. ny) then


                    if ( any( plevelimage(source(i)%imin : source(i)%imax, source(i)%jmin : &
                                      source(i)%jmax) .eq. source(i)%number .and.           &
                          elevelimage(source(i)%imin : source(i)%imax, source(i)%jmin :     &
                            source(i)%jmax) .lt. 0 )) then                                       
                          
               
                        source(i)%qualityFlag = ibset(source(i)%qualityFlag, 8)  
                    end if
                end if    
            else

                ! Check for any point of an extended-source lying on image boundary
                if ( source(i)%imin .eq. 0 .or. source(i)%imax .eq. nx .or. &
                     source(i)%jmin .eq. 0 .or. source(i)%jmax .eq. ny) then

                    source(i)%qualityFlag = ibset(source(i)%qualityFlag, 7 )  

                     
                end if
         
 

            end if

        end do

    end if


end subroutine SetSourceFlags
!***********************************************************************
!
!**********************************************************************
subroutine setCentralRegionFlags(x, y, source, n)
    real(single) :: x(:), y(:)
    type(sourceParameters), pointer :: source(:)
    integer(int16), intent(in) :: n
    real(single), parameter :: xc = 992.0, yc = 1063.0, radius = 160.0
    real(single) :: dx, dy
    integer(int16) :: i


    do i = 1, n
     
        if (x(i) .lt. (xc - radius) .or. x(i) .gt. (xc + radius) .or. &
            y(i) .lt. (yc - radius) .or. y(i) .gt. (yc + radius)) cycle
        dx = x(i) - xc
        dy = y(i) - yc
        if (sqrt(dx * dx + dy * dy) .lt. radius) then
            source(i)%qualityFlag = ibset(source(i)%qualityFlag, 5)  ! BIT 5 set for sources in central enhanced region
        end if
    end do

end subroutine setCentralRegionFlags

!***********************************************************************
!
!**********************************************************************
subroutine setCornerFlags(x, y, source, n)
    real(single) :: x(:), y(:)
    type(sourceParameters), pointer :: source(:)
    integer(int16), intent(in) :: n
   
  
    integer(int16) :: i 
    real(single), parameter :: xcentre = 979.2, ycentre = 1016.0
    real(single), parameter :: semia = 1300.0, semib = 1300.0     
    real(single), parameter :: theta = 90.0 * degreesToRadians
    integer :: noCornerFlagsSet

!  write(*, *) "Setcornerflags"




    noCornerFlagsSet = 0
    do i = 1, n
     
      if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then
    
!                write(*, *) "omsourceflags line 249 - source ", i, " x=", source(i)%xpos, &
!                     " y=", source(i)%ypos, " number=", source(i)%number, " set quality flag bit 7"
   
!               write(*, *) "x=", x(i), " y=", y(i)
                source(i)%qualityFlag = ibset(source(i)%qualityFlag, 7)
                noCornerFlagsSet = noCornerFlagsSet + 1
            end if

    end do


    write(errstr,'(a, i6)') "No of corner flags set=", noCornerFlagsSet
    call message(AppMsg, Verbose)


end subroutine setCornerFlags



 
!*********************************************
!
!
!********************************************
Subroutine LookForReadOutStreaks(source, image)
    type(sourceParameters), pointer :: source(:)

    real(single), pointer, dimension(:, :) :: image
    

    integer :: nMin, i0, i, n1, n2, n3, n4, n5, n6
    integer :: nx, ny, nStreaks
    real(single) :: width
  

    if (.not. associated(source)) return

    write(errstr,'(a)') "Checking for read-out streaks"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    nMin = int( 0.52 * (ny + 1)) 
    width = real(10 / binx)
    nStreaks = 0  
    i0 = 0
    do i = 0, nx

        if ( i0 < 3) then 

            
            n4 = count(image(i0, :) > image(i0 + 1, :) ) 
            n5 = count(image(i0, :) > image(i0 + 2, :) ) 
            n6 = count(image(i0, :) > image(i0 + 3, :) )
            n1 = count(image(i0, :) > image(i0 + 4, :) ) 
            n2 = count(image(i0, :) > image(i0 + 5, :) ) 
            n3 = count(image(i0, :) > image(i0 + 6, :) )
!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6

        else if ( i0 > (nx - 5)) then


            n1 = count(image(i0, :) > image(i0 - 1, :) ) 
            n2 = count(image(i0, :) > image(i0 - 2, :) ) 
            n3 = count(image(i0, :) > image(i0 - 3, :) )
                      

            n4 = count(image(i0, :) > image(i0 - 4, :) ) 
            n5 = count(image(i0, :) > image(i0 - 5, :) ) 
            n6 = count(image(i0, :) > image(i0 - 6, :) )

!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6

        else

            n1 = count(image(i0, :) > image(i0 - 1, :) ) 
            n2 = count(image(i0, :) > image(i0 - 2, :) ) 
            n3 = count(image(i0, :) > image(i0 - 3, :) )
            i0 = i0 + 1
            n4 = count(image(i0, :) > image(i0 + 1, :) ) 
            n5 = count(image(i0, :) > image(i0 + 2, :) ) 
            n6 = count(image(i0, :) > image(i0 + 3, :) )

!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6


        end if
        if (n1 > nMin .and. n2 > nMin .and. n3 > nMin .and. n4 > nMin &
                      .and. n5 > nMin .and. n6 > nMin) then
            write(errstr,'(a, i6, a, f7.2, a, f7.2)') &
                      "Possible read-out streak centred on column ", i0
            call message(AppMsg, Verbose)

            where(abs(source%xpos - real(i0) + 1.0) .lt. width &
                 .and. source%significance .lt. 50.0) source%qualityFlag = ibset(source%qualityFlag, 1)  ! BIT 1 set for sources on read-out streaks

            nStreaks = nStreaks + 1
            i0 = i0 + 10
          
        else
            i0 = i0 + 1
        end if
        if ( (nx - i0) < 5) exit
  

    end do

    write(errstr,'(a, i5)') "Number of possible read-out streaks found=", nstreaks

    call message(AppMsg, Verbose)

    write(errstr,'(a, i5)') "Number of source flags set=", count(source%qualityFlag .eq. 2 .or. source%qualityFlag .eq. 3)

    call message(AppMsg, Verbose)

end Subroutine LookForReadOutStreaks

!*********************************************
!
!
!*********************************************
Subroutine LookForStarSpikes(source, image, nsize)
    type(sourceParameters), pointer :: source(:)

    real(single), pointer :: image(:, :)
    integer, intent(in) :: nsize

    integer :: i, j, k, l, n, n1, n2, n3, n4
    integer :: nx, ny, nSpikes
    real(single) :: xl, xu, yl, yu
    real(single) :: xcentre, ycentre, xpos, ypos, theta
    real(single) :: radius, dr, cs, sn, dx, dy


    real(single) :: f1(nsize), f2(nsize), f3(nsize)
    real(single) :: tol
    integer :: nsize1, nsize2, nMin
    integer :: rMin, rMax

    if (.not. associated(source)) return

    n = size(source)

    write(errstr,'(a)') "Checking for stars with diffraction spikes"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    nSpikes = 0

    xl = 0.5
    yl = 0.5
    xu = size(image, 1) - 0.5
    yu = size(image, 2) - 0.5


    nsize1 = nsize / 2
    nsize2 = nsize1 + 1

    nMin = nsize

    rmin = 30 / binx
    rMax = 80 / binx

    tol = real(8 / binx)
    l = 0
    dr = 1.0 / real(binx)
iLoop : do i = 1, n
        if (source(i)%rawCountRate .lt. 50.0 .and. .not. btest(source(i)%qualityFlag, 4)) cycle
    



        xcentre = source(i)%xpos - 0.5
        ycentre = source(i)%ypos - 0.5
!     write(*, *) "xcentre=", xcentre, " ycentre=", ycentre, " rmin=", rmin, " rmax=", rmax

        do k = rmin, rmax
!write(*, *) "rmin=", rmin, " rmax=", rmax


            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 45.0, real(k), dr, f1, f2, f3)
            n1 = count(f2 .gt. f1)
            n2 = count(f3 .gt. f1)


            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 45.0, -real(k), -dr, f1, f2, f3)


            n3 = count(f2 .gt. f1)
            n4 = count(f3 .gt. f1)

!write(*, *) "Line 2400 n1=", n1, " n2=", n2, " n3=", n3, &
!" n4=", n4, " nmin=", nmin, " nsize=", nsize, " nsize1", nsize1, " nsize2=", nsize2

            if (n1 .ge. nmin) l = l + 1
            if (n2 .ge. nmin) l = l + 1
            if (n3 .ge. nmin) l = l + 1
            if (n4 .ge. nmin) l = l + 1


            call GetStarSpikeCounts(image, nsize1, xcentre, ycentre, 135.0, real(k), dr, f1, f2, f3)

            n1 = count(f2 .gt. f1)
            n2 = count(f3 .gt. f1)



            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 135.0, -real(k), -dr, f1, f2, f3)



            n3 = count(f2 .gt. f1)
            n4 = count(f3 .gt. f1)

            if (n1 .ge. nmin) l = l + 1
            if (n2 .ge. nmin) l = l + 1
            if (n3 .ge. nmin) l = l + 1
            if (n4 .ge. nmin) l = l + 1


!write(*, *) "Line 2438 n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, &
!" nmin=", nmin, " nsize=", nsize, " nsize", nsize1, " nsize2=", nsize2

            if (l .gt. 7) then

                nspikes = nspikes + 1

                write(errstr,'(a, i5, a, f6.2, a, f6.2)') &
                        "Possible diffraction spikes found associated with star ", &
                  i, " x=", source(i)%xpos, " y=", source(i)%ypos

                call message(AppMsg, Verbose)

                !*******************************************
                ! Flag sources near spikes
                !*******************************************

                theta = 45.0 * degreesToRadians
                sn = sin(theta)
                cs = cos(theta)

                rmin = -180 / binx

                do j = 1, 361

                    radius = real(rmin) + float(j - 1) * dr

                    xpos = xcentre + radius * cs
                    if (xpos .lt. xl .or. xpos .gt. xu) cycle

                    ypos = ycentre + radius * sn
                    if (ypos .lt. yl .or. ypos .gt. yu) cycle

                    do l = 1, n
                        if (l .eq. i) cycle
                        if (source(l)%qualityFlag .ge. 8) cycle
                        if (source(l)%rawCountrate .ge. 20.0) cycle

                        dx = source(l)%xpos - xpos

                        dy = source(l)%ypos - ypos
                        if (sqrt(dx * dx + dy * dy) .le. tol) then

                            source(l)%qualityFlag = ibset(source(l)%qualityFlag, 3)  ! BIT 3 set for sources on star spikes

!                            write(errstr,'(a, i5, a)') "Setting quality flag of source ", l,  " to 8"

!                            call message(AppMsg, Verbose)

                        end if
                    end do
                end do

                theta = 135.0 * degreesToRadians
                sn = sin(theta)
                cs = cos(theta)

                do j = 1, 361

                    radius = real(rmin) + float(j - 1) * dr

                    xpos = xcentre + radius * cs
                    if (xpos .lt. xl .or. xpos .gt. xu) cycle

                    ypos = ycentre + radius * sn
                    if (ypos .lt. yl .or. ypos .gt. yu) cycle

                    do l = 1, n
                        if (l .eq. i) cycle
                        if (source(l)%qualityFlag .ge. 8) cycle
                        if (source(l)%rawCountrate .ge. 20.0) cycle

                        dx = source(l)%xpos - xpos
                        dy = source(l)%ypos - ypos
                        if (sqrt(dx * dx + dy * dy) .le. tol) then

                            source(l)%qualityFlag = ibset(source(l)%qualityFlag, 3)  ! BIT 3 set for sources on star spikes

!                            write(errstr,'(a, i5, a)') "Setting quality flag of source ", l,  " to 8"

!                            call message(AppMsg, Verbose)

                        end if
                    end do
                end do

                cycle iLoop

            end if
        end do
    end do iLoop

    write(errstr,'(a, i5)') "Number of possible diffraction-spikes found=", nspikes

    call message(AppMsg, Verbose)

    write(errstr,'(a, i5)') "Number of source flags set=", count(source%qualityFlag .ge. 8)

    call message(AppMsg, Verbose)

end Subroutine LookForStarSpikes

!**************************************************************************************
!
!
!*************************************************************************************

subroutine GetStarSpikeCounts(image, nsize1, xcentre, ycentre, angle, startRadius, dr, f1, f2, f3)
    real(single), pointer :: image(:, :)
    integer, intent(in) :: nsize1

    real(single), intent(in) :: xcentre, ycentre, startRadius, angle, dr
    real(single), dimension(1 : nsize1) :: f1, f2, f3
    real(single) :: radius, cs, sn, theta, xpos, ypos, xpos1, ypos1
    real(single) :: dx, dy, xl, yl, xu, yu
    integer :: j

    theta = angle * degreesToRadians

    sn = sin(theta)
    cs = cos(theta)

    dx = real(8 / binx)
    dy = real(8 / binx)

    xl = 0.5
    yl = 0.5
    xu = size(image, 1) - 0.5
    yu = size(image, 2) - 0.5

    f1 = 0.0
    f2 = 0.0
    f3 = 0.0

    do j = 1, nsize1
        radius = startRadius + real(j - 1) * dr

        xpos = xcentre + radius * cs
        if (xpos .lt. xl .or. xpos .gt. xu) cycle

        ypos = ycentre + radius * sn
        if (ypos .lt. yl .or. ypos .gt. yu) cycle

        f1(j) = bilinearinterpolate(image, xPos, yPos)
        xpos1 = xpos - dx * sn
        if (xpos1 .lt. xl .or. xpos1 .gt. xu) cycle

        ypos1 = ypos + dy * cs
        if (ypos1 .lt. yl .or. ypos1 .gt. yu) cycle

        f2(j) = bilinearinterpolate(image, xPos1, yPos1)
        xpos1 = xpos + dx * sn
        if (xpos1 .lt. xl .or. xpos1 .gt. xu) cycle

        ypos1 = ypos - dy * cs
        if (ypos .lt. yl .or. ypos .gt. yu) cycle

        f3(j) = bilinearinterpolate(image, xPos1, yPos1)


    end do

end subroutine GetStarSpikeCounts

!************************************************************
!  Converts source-list file coordinates to full-frame
!  coordinates
!
!************************************************************
subroutine ConvertCoordsToFullFrame1(xCoord, yCoord, il, iu, sourceListDataSetH)

    real(single) :: xCoord(:), yCoord(:)
    integer, intent(in) :: il, iu
    type(dataSetT) :: sourceListDataSetH
 
    integer(int32) :: xbin, ybin
    logical(boolean) :: binBpe
 
    character(100) :: windowString
    character(1) :: c
    integer :: windowX0, windowY0, windowX1, windowY1, winDx, winDy

    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    if (hasAttribute(sourceListDataSetH, "DETSEC")) then
        windowString = stringAttribute(sourceListDataSetH, "DETSEC")
        read (windowString, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  c, windowX0, c, windowX1, c,  windowY0, c, windowY1, c
       
        winDX = windowX1 - windowX0 + 1
        winDY = windowY1 - windowY0 + 1
           
      
    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")
 
    ! Calculate the image binning.
    xBin = 2 ** xbin
    yBin = 2 ** ybin


    if (binBPE) then
        xBin = xBin * 2
        yBin = yBin * 2
    end if

    windowX0 = windowX0 * xBin
    windowY0 = windowY0 * yBin


    call ConvertCoordsToFullFrame2(xCoord(il : iu), yCoord(il : iu), &
                windowX0, windowY0, xBin, yBin)


end subroutine ConvertCoordsToFullFrame1



!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsToFullFrame2(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single), dimension(:) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin
    integer :: i

    ! Convert coordinates to full-frame unbinned coordinates  
    do i = 1, size(xCoord)

        xCoord(i) = xCoord(i) * real(xBin) + real(x0)
        yCoord(i) = yCoord(i) * real(yBin) + real(y0)

    end do



end subroutine ConvertCoordsToFullFrame2
!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsToFullFrame3(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single), intent(inout) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin


    xCoord = xCoord * real(xBin) + real(x0)
    yCoord = yCoord * real(yBin) + real(y0)

end subroutine ConvertCoordsToFullFrame3
!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsFromFullFrame(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single), intent(inout) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin


    xCoord = (xCoord - real(x0)) / real(xbin)
    yCoord = (yCoord - real(y0)) / real(yBin)

end subroutine ConvertCoordsFromFullFrame






!***************************************************************
!
!
!**************************************************************
subroutine setSmokeRings(nSources, source)
  
    integer(int16), intent(in) :: nSources
    type(sourceParameters), pointer :: source(:)

 
    integer :: i, j
    real(single) :: dx, dy, radius
    real(single) :: width
    if (nSources == 0) return
     
    width = real(40 / binx)


!write(*, *) "Setsmokerings line 935"

    do i = 1, nSources
        if ( btest(source(i)%qualityFlag, 2) ) then ! Then has been detected as a smoke ring 

!            write(*, *) "Checking source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos  

            do j = i - 1, 1, -1
                if ( btest(source(i)%qualityFlag, 4) ) cycle
              
        
                dx = source(i)%xpos - source(j)%xpos
                if (dx > width) cycle
        
                dy = source(i)%ypos - source(j)%ypos
           
                if (dy > width) cycle
        
                radius = sqrt(dx * dx + dy * dy)

                if (radius .le. width) then

!  write(*, *) "Setting quality flag on source ", j, " x=", source(j)%xpos, " y=", source(j)%ypos
                    ! Remember that it is probably a smoke ring     
                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 2) 
  
                end if
 
            end do 


            do j = i + 1, nSources
                if ( btest(source(j)%qualityFlag, 4) ) cycle 
                dx = source(i)%xpos - source(j)%xpos
                if (dx > width) cycle
              
                dy = source(i)%ypos - source(j)%ypos
        
                if (dy > width) cycle
        
                radius = sqrt(dx * dx + dy * dy)

                if (radius .le. width) then

                    ! Remember that it is probably a smoke ring     
                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 2) 
!    write(*, *) "Setting quality flag on source ", j, " x=", source(j)%xpos, " y=", source(j)%ypos
                    

                end if
 
            end do 

        end if
    end do

end subroutine setSmokeRings
!***************************************************
! Subroutine 
!
!***************************************************
Subroutine ReadOutStreaks1(n, m, jIndex,   &
                           countTable,     &
                           x,              &
                           extensionFlag,  &
                           qualityFlag)
    integer, intent(in) :: n, m, jIndex(:)
    real(single) :: countTable(:, :), x(:)
    integer(int8) :: extensionFlag(:)
    integer(int16):: qualityFlag(:)
    real(single), pointer :: counts(:) => null()
    integer :: i, j ,k

    counts => allocateSingle(1, n, "counts")


    do i = 1, n
        counts(i) = -1.0
        do j = 1, m
            k = jIndex(j)
            if (countTable(i, k) .gt. 0.0 .and. countTable(i, k) .lt. 10000.0) then
                counts(i) = max(counts(i), countTable(i, k))
            end if
        end do
    end do      

    call ReadOutStreaks2(n,                &
                         counts,           &
                         x,                &
                         extensionFlag,    &
                         qualityFlag)
  
    deallocate(counts)
    nullify(counts)

end subroutine ReadOutStreaks1

!***************************************************
! Subroutine 
!
!***************************************************
Subroutine ReadOutStreaks2(n,              &
                  countTable,              &
                  x,                       &
                  extensionFlag,           &
                  qualityFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:)
    integer(int8) :: extensionFlag(:)
    integer(int16) :: qualityFlag(:)
    integer :: i, j, l
  
    real(single), pointer :: xStreak(:) => null(), streakhalfwidth(:) => null()
    integer, parameter :: maxNoStreaks = 100
    integer :: noStreaks, noFlagsSet
      
    real(single), parameter :: halfWidth1 = 16.0  ! IE 12 pixel aperture radius + 7.0 pixel streak width
    real(single), parameter :: halfWidth2 = 19.0  ! IE 12 pixel aperture radius + 7.0 pixel streak width
    real(single), parameter :: constant = 70.0


    xStreak => allocateSingle(1, maxNoStreaks, "xStreak")
    streakhalfwidth => allocateSingle(1, maxNoStreaks, "streakhalfwidth")
 

    noStreaks = 0
!write(*, *) "n=", n, " no of pointsource flags=", count(extensionFlag .eq. 0), " extended=", count(extensionFlag .eq. 1)  
iLoop :  do i = 1, n  
!write(*, *) "Source ", i, " counttble=", countTable(i), " constant=", constant, " exposure=", exposure(i), &
!" extension flag=", extensionFlag(i)

            if (extensionFlag(i) .eq. 1) cycle
                   
            if ( countTable(i) .gt. 0.0 .and. countTable(i) .lt. 10000.0) then

                  if ( btest(qualityFlag(i), 4) .or. countTable(i) .ge. constant) then
  


                    do l = 1, noStreaks
                        if (abs(x(i) - xstreak(l)) .lt. 5.0) cycle iLoop
                    end do

!write(*, *) "counttble=", countTable(i), " constant=", constant, " exposure=", exposure
                    noStreaks = noStreaks + 1
                    xStreak(noStreaks) = x(i)
                    write(errstr,'(a, f8.2)') "Possible read-out streak found centred on pixel ", x(i)
                    call message(AppMsg, Verbose)

         
                    if (countTable(i) .lt. 220.0) then
                        streakhalfwidth(noStreaks) = halfWidth1
                    else
                        streakhalfwidth(noStreaks) = halfWidth2
                    end if
       
 
                    if (noStreaks .eq. maxNostreaks) exit iLoop     
       
                end if

            end if

        
    end do iLoop

    write(errstr, '(a, i6)') "Number of read-out streaks found=", noStreaks
    call message(AppMsg, Verbose)
  
   
!write(*, *) "Readoutstreaks line 238 minval qualityflag=", minval(qualityFlag), " max=", maxval(qualityFlag)

!   return
    noFlagsSet = 0
    do i = 1, n
!        if (countTable(i) .lt. 0.0 .or. countTable(i) .gt. 10000.0) cycle

        do j = 1, noStreaks

!            if (btest(qualityFlag(i), 1)) cycle


            if (abs(x(i) - xstreak(j)) .lt. streakhalfWidth(j)) then

                qualityFlag(i) = ibset(qualityFlag(i), 1)  ! BIT 1 SET
                noFlagsSet = noFlagsSet + 1
                exit
            end if
        end do
    end do

    write(errstr,'(a, i6)') "Number of sources with quality-flag bit 1 set=", noFlagsSet
    call message(AppMsg, Verbose)


!write(*, *) "Readoutstreaks line 252 minval qualityflag=", minval(qualityFlag), " max=", maxval(qualityFlag)


    deallocate(xStreak, streakhalfwidth)
    nullify(xStreak, streakhalfwidth)    

end subroutine ReadOutStreaks2


!***************************************************
!  subroutine BrightSourceProximityFlags
!
!***************************************************
Subroutine  BrightSourceProximityFlags1(n,      &
                      countTable,              &
                      x, y,                    &
                      qualityFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:), y(:)
    integer(int16), dimension(:) :: qualityFlag
   
    integer :: i, j
    
  
    real(single) :: radius, dx, dy, maxrad, countsRatio
  
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 80.0    

    do i = 1, n

      
    
    
        if ( btest(qualityFlag(i), 4)   &
                  .or. countTable(i) .ge. minimumCountRate ) then
 
            maxRad = min(70.0, 35.0 * countTable(i) / minimumCountRate)
 
            do j = i - 1, 1, -1
                if ( btest(qualityFlag(j), 4) )  cycle
                if ( btest(qualityFlag(j), 6) )  cycle

  
                countsRatio = countTable(j) / max(1.0, countTable(i))
                if (countsRatio > 1.0) cycle 



                dx = x(i) - x(j)
                dy = y(i) - y(j)
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then

                    qualityFlag(j) = ibset(qualityFlag(j), 6)

                end if
                  
            end do 

            do j = i + 1, n
                if ( btest(qualityFlag(j), 4) )  cycle
                if ( btest(qualityFlag(j), 6) )  cycle
  
                countsRatio = countTable(j) / max(1.0, countTable(i))
                if (countsRatio > 1.0) cycle 

                dx = x(i) - x(j)
                dy = y(i) - y(j)
                   
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then
                    qualityFlag(j) = ibset(qualityFlag(j), 6)

   
                end if
    
            end do
        end if
    end do
    

end subroutine BrightSourceProximityFlags1 


!***************************************************
!  subroutine BrightSourceProximityFlags
!  For use when a possible exclusion radius is given for
!  the sources
!***************************************************
Subroutine  BrightSourceProximityFlags2(n,      &
                      countTable,              &
                      x, y,                    &
                      exclusionRadius,         & 
                      qualityFlag)
    integer, intent(in) :: n
    real(single), dimension(:) :: countTable, x, y, exclusionRadius
    integer(int16)  :: qualityFlag(:)
  
    integer :: i, j
    
  
    real(single) :: radius, dx, dy, maxrad, countsRatio
  
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 80.0    

    do i = 1, n
     
  
   
        if ( btest(qualityFlag(i), 4)   &
                  .or. countTable(i) .ge. minimumCountRate ) then
            
            maxRad = max(exclusionRadius(i), min( 70.0, 35.0 * countTable(i) / minimumCountRate))
 
           
            do j = i - 1, 1, -1
                if ( btest(qualityFlag(j), 4) )  cycle
                if ( btest(qualityFlag(j), 6) )  cycle
  
                countsRatio = countTable(j) / max(1.0, countTable(i))
                if (countsRatio > 1.0) cycle 



                dx = x(i) - x(j)
                dy = y(i) - y(j)
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then

                    qualityFlag(j) = ibset(qualityFlag(j), 6)
             
                end if
                  
            end do 

            do j = i + 1, n
                if ( btest(qualityFlag(j), 4) )  cycle
                if ( btest(qualityFlag(j), 6) )  cycle
  
                countsRatio = countTable(j) / max(1.0, countTable(i))
                if (countsRatio > 1.0) cycle 




                dx = x(i) - x(j)
                dy = y(i) - y(j)
                   
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then  
                    qualityFlag(j) = ibset(qualityFlag(j), 6)
                end if
    
            end do
        end if
    end do
    
  
end subroutine BrightSourceProximityFlags2 



!***************************************************
! Subroutine SmokeRings
!
!***************************************************
Subroutine SmokeRings1(n, m, jIndex,           &
                      countTable,              &
                      x, y,                    &
                      qualityFlag)
    integer, intent(in) :: n, m, jIndex(:)
    real(single) :: countTable(:, :), x(:), y(:)
    integer(int16) :: qualityFlag(:)

    real(single), pointer :: counts(:) => null()
    integer :: i, j ,k

    counts => allocateSingle(1, n, "counts")

    do i = 1, n
        counts(i) = -1.0
        do j = 1, m
            k = jIndex(j)
            if (countTable(i, k) .gt. 0.0 .and. countTable(i, k) .lt. 10000) then
                counts(i) = max(counts(i), countTable(i, k))
            end if
        end do
    end do      


    call SmokeRings2(n, counts,                &
                      x, y,                    &
                      qualityFlag)
   

    deallocate(counts)
    nullify(counts)
 
end subroutine SmokeRings1


!***************************************************
! Subroutine SmokeRings
!
!***************************************************
Subroutine SmokeRings2(n,                      &
                      countTable,              &
                      x, y,                    &
                      qualityFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:), y(:)
    integer(int16) :: qualityFlag(:)
    integer :: i, j
    real(double) :: xSmokeRing, ySmokeRing
    integer, parameter :: maxNoSmokeRings = 100
    integer :: noSmokeRings


    real(single), parameter :: halfWidth = 38.0 ! IE 26 ring radius + 12 photometry radus)
  
  
    real(double) :: rStar, rRing, theta


  
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 60.0 ! Was 70.0    


    if (n < 1) return

      
 

    noSmokeRings = 0

    ! Loop through the sources to find any that may produce a smoke-ring
iLoop :  do i = 1, n  


!write(*, *) "checking source ",  i,  " for smoke rings ", " x=", x(i), " y=", y(i), &
!" countrate=", countTable(i), " quality flag=", qualityFlag(i), " extension=", extensionFlag(i)
!write(*, *) "btest(qualityflag, 4)=",  btest(qualityFlag(i), 4)


        if (countTable(i) .gt. 0.0 .and. countTable(i) .lt. 10000.0) then
            if ( counttable(i) .ge. minimumCountRate .or. btest(qualityFlag(i), 4) ) then ! This star may have produced a smke ring
                if ( btest(qualityFlag(i), 2) ) then ! Then has been detected as a smoke ring 
                    qualityFlag(i) = ibclr(qualityFlag(i), 2)   ! BIT 2 UNSET
               
                    cycle
                end if
!write(*, *) "checking source ",  i,  " for smoke rings ", " x=", x(i), " y=", y(i), " count rate=", counttable(i)
!write(*, *) "QualityFlag=", qualityFlag(i), " extensionflag=", extensionflag(i)


                ! The source is predicted to produce a smoke-ring 
                ! Predict where smoke-ring is
                ! Radial distance of star from centre of full-frame image
                ! xCentre and yCentre are defined in omconstants_mod (1024.5. 1024.5)
                rStar = RadialDistance( xCentre, yCentre, dble(x(i)), dble(y(i)))
                theta = PositionAngle( xCentre, yCentre, dble(x(i)), dble(y(i)))


               



 !               rRing = 1.206 * rStar


                xsmokeRing = smokeRingXPosition(x(i), y(i))

                ysmokeRing = smokeRingYPosition(x(i), y(i))

                rRing = RadialDistance( xCentre, yCentre, xsmokeRing, ySmokeRing)
         
 
                if (rRing .lt. 10) cycle 
            
                noSmokeRings = noSmokeRings + 1
              
 !               xSmokeRing = xCentre + rRing * cos(theta)
 !               ySmokeRing = yCentre + rRing * sin(theta)
                        
            


                do j = 1, n
                    if (i .eq. j) cycle
                         
                    if ( RadialDistance(dble(x(j)), dble(y(j)), xSmokeRing, ySmokeRing) .le. halfwidth) then
                        qualityFlag(j) = ibset(qualityFlag(j), 2)   ! BIT 2 SET
                
!                    write(*, *) "set source ", j, " x=", x(j), " y", y(j)        
                    end if

                end do
 
            end if

        end if
   
    end do iLoop

    write(errstr, '(a, i6)') "Number of smoke rings=", noSmokeRings
    call message(AppMsg, Verbose)
  
  

   
end subroutine SmokeRings2
!***************************************************************
!
!**************************************************************

real(single) function smokeRingXPosition(xStar, yStar)
    real(single), intent(in) :: xStar, yStar

    real(double), parameter :: a1 = 9.6950_double
    real(double), parameter :: b1 = 1.2052_double
    real(double), parameter :: c1 = 1.1027d-03
    real(double), parameter :: d1 = 2.9117d-06
    real(double), parameter :: e1 = 1.8790d-06
    real(double), parameter :: f1 = 4.2347d-07

    real(double) :: x, y

    x = dble(xStar - xCentre)
    y = dble(yStar - yCentre)


    smokeRingXPosition = real(a1 + b1 * x + c1 * y + d1 * x * x + &
                         e1 * y * y + f1 * x * y) + xCentre


end function smokeRingXPosition


!***************************************************************
!
!**************************************************************

real(single) function smokeRingYPosition(xStar, yStar)
    real(single), intent(in) :: xStar, yStar

    real(double), parameter :: a2 = -4.8165_double
    real(double), parameter :: b2 = 5.0884d-04
    real(double), parameter :: c2 = 1.2071_double
    real(double), parameter :: d2 = -2.8810d-07
    real(double), parameter :: e2 = 1.549d-06
    real(double), parameter :: f2 = -1.0783d-06

    real(double) :: x, y

    x = dble(xStar - xCentre)
    y = dble(yStar - yCentre)


    smokeRingYPosition = real(a2 + b2 * x + c2 * y + d2 * x * x + &
                         e2 * y * y + f2 * x * y) + yCentre


end function smokeRingYPosition





!************************************************
! 
!***********************************************
real(double) Function RadialDistance(x1, y1, x2, y2)
    real(double), intent(in) :: x1, y1, x2, y2
    real(double) ::  dx, dy

    dx = x2 - x1
    dy = y2 - y1
    RadialDistance = dsqrt(dx * dx + dy * dy)

end function RadialDistance
!************************************************
! 
!***********************************************
real(double) Function PositionAngle(x1, y1, x2, y2)
    real(double), intent(in) :: x1, y1, x2, y2
    real(double) ::  dx, dy

    dx = x2 - x1
    dy = y2 - y1
    if (dx .eq. 0.0_double .and. dy .eq. 0.0_double) then
        PositionAngle = 0.0_double
        return
    end if

   PositionAngle = datan2(dy, dx)

end function PositionAngle

!***********************************************
!
!***********************************************
subroutine EdgeSources(n, source)
                      
    integer, intent(in) :: n
    type(sourceparameters), pointer :: source(:)

    integer :: i
    real(single), parameter :: minDist = 12.0
    real(single), parameter :: width = 2049.0
 
    integer :: count

!write(*, *) "edgesources line 684 minval source%qualityflag=", minval(source%qualityFlag), &
!" max=", maxval(source%qualityFlag)

    write(errstr, *) "Checking for sources near the edge of the image"

    call message(AppMsg, Verbose)

    count = 0
    do i = 1, n
!write(*, *) i, " x=", source(i)%xpos, " y=", source(i)%ypos, " flag=", source(i)%qualityFlag

        if (source(i)%xpos .lt. minDist .or. (width - source(i)%xpos).lt. minDist .or. &
            source(i)%ypos .lt. minDist .or. (width - source(i)%ypos).lt. minDist) then


  

 !                write(*, *) "omsourceflags line 1626 edge sources - - source ", i, " x=", source(i)%xpos, &
 !                    " y=", source(i)%ypos, " number=", source(i)%number, " set confusion flag to 7"
       



            source(i)%qualityFlag = ibset(source(i)%qualityFlag, 7)  

            count = count + 1
        end if

        !**************************************************************************
        ! Check for sources close to a corner
        !*************************************************************************
       
    end do

    write(errstr, *) "Number of source flags set=", count
    call message(AppMsg, Verbose)
    
!write(*, *) "edgesources line 720 minval source%qualityflag=", minval(source%qualityFlag), &
!" max=", maxval(source%qualityFlag)


end subroutine EdgeSources

!***************************************************
! Subroutine StarSpikes
!
!***************************************************
Subroutine StarSpikes(n, source, xcoord, ycoord)
   
                      
    integer, intent(in) :: n
    type(sourceparameters), pointer :: source(:)
    real(single), dimension(:) :: xcoord, ycoord
    integer(int16) :: i, j, k, l, m
    real(single) :: dx, dy
    type(sourceparameters) :: tempsource
    type(sourceparameters), pointer :: tempsources(:) => null()
  
    integer(int8) :: maxCount, count1
    integer(int16) :: count2, count3
            
    real(single), parameter :: angle(4) = (/45.0, 225.0, 135.0, 315.0/)   
    integer(int16), parameter :: noOffsets = 13


    real(single), parameter :: offset(noOffsets) = (/0.0, -0.5, 0.5, &
                                                    -1.0, 1.0, -1.5, 1.5, -2.0, 2.0, &
                                                    -2.5, 2.5, -3.0, 3.0/)

!integer, save :: call = 0
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 70.0    

 
    integer(int16) :: noSpikes, minRadius, maxRadius, minrad1, maxrad1
    logical, pointer :: useStar(:) => null()  
    integer, pointer :: index(:) => null()
    integer :: il, iu
    real(single) :: minRate, xc, yc

    call StarSpikes1(n, source, xcoord, ycoord)
    return

    minRate = 5.0

!if (call > 10) return
!call = call + 1

    write(errstr, *) "Checking for stars with diffraction spikes"
    call message(AppMsg, Verbose)

              
    noSpikes = 0
    allocate(useStar(1 : n))

    allocate(tempsources(1 : n))
    tempSources = source(1 : n)

  
    index => allocateInt32(1, n, "index")
    call indexSort(xcoord, index)

    do i = 1, n
        k = index(i)
        tempSources(i) = source( k )
        tempSources(i)%xpos = xcoord( k )
        tempSources(i)%ypos = ycoord( k )


!write(*, *) i, " ) theta=", tempSources(i)%theta
       
    end do

   

iLoop : do i = 1, n  
        
!if (index(i) .ne. 261) cycle

!write(*, *) i, ") count rate=", tempsources(i)%rawCountRate, " exposure=", tempSources(i)%exposure, &
!  " x=", tempSources(i)%xpos, " y=", tempsources(i)%ypos, " qualityflag=", tempsources(i)%qualityFlag             

        if ( tempsources(i)%rawCountRate .ge. minimumCountRate ) then ! This star may have produced spikes
               

            write(*, *) "Checking source ", index(i), " x=", tempSources(i)%xpos, " y=", tempSources(i)%ypos 

            !*******************************************************************************
            ! For position angles 45 and 135  check how many stars lie along a radial line
            ! away from the centre of the star
            !*******************************************************************************
            maxCount = 1
            dx = 0.0
            dy = 0.0

                    
            count2 = 0                       
            do j = 1, noOffsets    

                tempsource%xpos = tempSources(i)%xpos + offset(j)           
              
                do k = 1, noOffsets
    
                    minRadius = 1000
                    maxRadius = 0
        
                    count1 = 0
                    tempsource%ypos = tempSources(i)%ypos + offset(k) 
                    usestar = .true.
                    count3 = 0

                    do l = 1, 4

                              
                        !******************************************************************
                        ! Check for stars along position angle 45 and 135 degrees
                        !***************************************************************** 
                        if (CheckForStars(tempSource, tempsources, i, angle(l), &
                                 useStar, minRadius, maxRadius, count3, minRate)) then
                            count1 = count1 + 1


                        end if
                    end do


!write(*, *) "Line 1696 maxount=", maxcount, " count1=", count1, " count2=", count2, " count3=", count3

                    if (count1 .gt. maxcount .or. (count1 == maxcount .and. count3 > count2)) then     

                        maxCount = count1        
                        count2 = count3
                        dx = offset(j)
                        dy = offset(k)
                                
                    end if         
                      
!                 write(*, *) "Line 1278 angl=", angle(l), " count2=", count2, " count1=", count1, " maxcount=", maxcount
  
                end do
            end do    
            


            if (maxcount .gt. 1) then
                noSpikes = noSpikes + 1                    
                write(errstr,'(a, i5, a, f7.2, a, f7.2, a, f8.2, a, f8.2, a, f8.2)') &
                               "Possible diffraction spikes found associated with star ", &
                                index(i), " x=", tempsources(i)%xpos, " y=", tempsources(i)%ypos, &
                                " raw-rate=", tempSources(i)%rawCountRate

                call message(AppMsg, Verbose)
              
!write(*, *) "Line 1131 minrad1=", minrad1, " maxrad1=", maxrad1, " dx=", dx, " dy=", dy, " maxcount=", maxcount
                minrad1 = 10
                maxrad1 = 120
                xc = tempSources(i)%xpos + dx
                yc = tempSources(i)%ypos + dy
           
                il = 1
                il = locate(n, tempSources%xpos, max(0.0, xc - real(maxrad1 - 5)), il)
           
                iu = il
                iu = locate(n, tempSources%xpos, xc + real(maxrad1 + 5), iu)
       
                usestar = .true.

                ! Set the flags
                do m = 1, 4
!                    call SetStarSpikeFlags(xc, yc, tempsources, i, il, iu, angle(m), minRad1, maxRad1, usestar) 
                end do




!call fatal("stopped")

            end if 
           
                            
        end if


    end do iLoop    
    write(errstr,'(a, i4)') "No of possible sources with diffraction-spikes=", noSpikes
    call message(AppMsg, Verbose)

    do i = 1, n
        k = index(i)
        tempSource = source( k )
        source( k ) = tempSources(i)
        source( k )%xpos = tempSource%xpos
        source( k )%ypos = tempSource%ypos

    end do


    deallocate(index, tempSources, usestar)
    nullify(index, tempSources, usestar)
    
    
end subroutine StarSpikes



!***************************************************
! Subroutine StarSpikes
!
!***************************************************
Subroutine StarSpikes1(n, source, xcoord, ycoord)
   
                      
    integer, intent(in) :: n
    type(sourceparameters), pointer :: source(:)
    real(single), dimension(:) :: xcoord, ycoord
    integer(int16) :: i, k, l, m
    
  
    type(sourceparameters), pointer :: tempsources(:) => null()
   
    type(sourceparameters) :: tempsource

    integer(int8) :: maxCount, count1
    integer(int16) :: count2, count3
            
    real(single), parameter :: angle(4) = (/45.0, 225.0, 135.0, 315.0/)   
    real(single), dimension(4) :: cosTheta, sintheta
    real(single) :: totalArea, SearchArea 
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 70.0, maxDistance = 5.0    

 
    integer(int16) :: noSpikes, minRadius, maxRadius
    logical, pointer :: useStar(:) => null()  
    integer, pointer :: index(:) => null()
    integer :: il, iu
    real(single) :: minRate, xc, yc, xl, xu, yl, yu
   
    do i = 1, 4
        cosTheta(i) = cos(angle(i) * degreesToRadians)
        sinTheta(i) = sin(angle(i) * degreesToRadians)
    end do

    minRate = 5.0

!if (call > 10) return
!call = call + 1

    write(errstr, *) "Checking for stars with diffraction spikes"
    call message(AppMsg, Verbose)

              
    noSpikes = 0
    allocate(useStar(1 : n))

    allocate(tempsources(1 : n))
    tempSources = source(1 : n)

  
    index => allocateInt32(1, n, "index")
    call indexSort(xcoord, index)

    do i = 1, n
        k = index(i)
        tempSources(i) = source( k )
        tempSources(i)%xpos = xcoord( k )
        tempSources(i)%ypos = ycoord( k )


!write(*, *) i, " ) theta=", tempSources(i)%theta
       
    end do

   

    do i = 1, n  
        
!if (index(i) .ne. 968) cycle

!write(*, *) i, ") count rate=", tempsources(i)%rawCountRate, " exposure=", tempSources(i)%exposure, &
!  " x=", tempSources(i)%xpos, " y=", tempsources(i)%ypos, " qualityflag=", tempsources(i)%qualityFlag             


        if ( tempsources(i)%rawCountRate .ge. minimumCountRate .or. btest(tempSources(i)%qualityFlag, 4) ) then ! This star may have produced spikes
               
  
   
!            write(*, *) "Checking source ", index(i), " x=", tempSources(i)%xpos, " y=", tempSources(i)%ypos 

            !*******************************************************************************
            ! For position angles 45 and 135  check how many stars lie along a radial line
            ! away from the centre of the star
            !*******************************************************************************
            maxCount = 1                    
            count2 = 0                       
            count1 = 0
                 
            usestar = .true.
            count3 = 0

       
            tempSource = tempSources(i)
            searchArea = 0.0

            xc = tempsource%xpos
            yc = tempsource%ypos

            do l = 1, 4
                


   
                minRadius = 10
                maxRadius = 120
        
      
              
                !******************************************************************
                ! Check for stars along position angle 45 and 135 degrees
                !***************************************************************** 
                if( CountStars(tempSource, tempsources, i, costheta(l), sintheta(l), &
                    useStar, minRadius, maxRadius, count3, minRate, maxDistance)) then
                    count1 = count1 + 1

                end if
                SearchArea = searchArea + real(maxRadius - minradius) * 2.0 * maxDistance 

                !*****************************************************************
                ! Obtain the maximum size of the search box
                !***************************************************************** 

                

            end do

            

            xl = max( xc - 150, 0.5 )
            yl = max( yc - 150.0, 0.5 )
            xu = min( xc + 150.0, 2048.5 )
            yu = min( yc + 150.0, 2048.5 )




!write(*, *) "Starspikes1 count1=", count1              
            if ( count1 .gt. 1) then

                !**************************************************************************
                ! Find the expected number of stars along the spikes
                !**************************************************************************
                count2 = NumberOfStarsInBox(tempSources, i, minrate, xl, xu, yl, yu)
                totalArea = (xu - xl) * (yu - yl)

!write(*, *) "Count2=", count2, " totalarea=", totalarea, " searcharea=", searcharea

                !****************************************************************************
                ! See if the number of star along the spikes is > 3 times the expected number
                !**************************************************************************** 
                if( real(count3) > 3.0 * real(count2) * SearchArea / totalArea) then

                    !*************************************************************
                    ! May have found diffraction spikes
                    !*************************************************************



                    noSpikes = noSpikes + 1                    
                    write(errstr,'(a, i5, a, f7.2, a, f7.2, a, f8.2, a, f8.2, a, f8.2)') &
                               "Possible diffraction spikes found associated with star ", &
                                index(i), " x=", tempsources(i)%xpos, " y=", tempsources(i)%ypos, &
                                " raw-rate=", tempSources(i)%rawCountRate

                    call message(AppMsg, Verbose)
              

                    minradius = 10
                    maxradius = 120
                    xc = tempSources(i)%xpos
                    yc = tempSources(i)%ypos
           
                    il = 1
                    il = locate(n, tempSources%xpos, max(0.0, xc - real(maxradius - 5)), il)
           
                    iu = il
                    iu = locate(n, tempSources%xpos, xc + real(maxradius + 5), iu)
       
                    usestar = .true.

                    ! Set the flags
                    do m = 1, 4


                        call SetStarSpikeFlags(xc, yc, tempsources, i, il, iu, &
                                costheta(m), sintheta(m), minRadius, maxRadius, usestar) 

                    end do
                end if
            end if           
                            
        end if


    end do     
    write(errstr,'(a, i4)') "No of possible sources with diffraction-spikes=", noSpikes
    call message(AppMsg, Verbose)




    do i = 1, n
        k = index(i)
        tempSource = source( k )
        source( k ) = tempSources(i)
        source( k )%xpos = tempSource%xpos
        source( k )%ypos = tempSource%ypos

    end do


    deallocate(index, tempSources, usestar)
    nullify(index, tempSources, usestar)
    
    
end subroutine StarSpikes1

!***************************************************************************
!
!**************************************************************************
Integer(int16) Function NumberOfStarsInBox(source, i, minrate, xl, xu, yl, yu)
      
    type(sourceparameters), pointer :: source(:)

   
    integer(int16), intent(in) :: i
    real(single), intent(in) :: minRate
    real(single), intent(in) :: xl, xu, yl, yu
    integer(int32) :: j, il, iu, n

   
    n = size(source)

    il = 1
    il = locate(n, source%xpos, xl - 2.0, il)
           
    iu = il
    iu = locate(n, source%xpos, xu + 2.0, iu)

    NumberOfStarsInBox = 0

    do j = il, iu
 
        if (j .eq. i) cycle    
        if(source(j)%ypos < yl) cycle
        if(source(j)%ypos > yu) cycle
     

        if ( (source(j)%rawCountRate > minRate .and. source(j)%rawCountRate < 999.0) &
                     .and. .not. isNullValue(source(j)%rawCountRate)) then
               
            cycle
        end if
        NumberOfStarsInBox = NumberOfStarsInBox + 1

    end do

!write(*, *) "Numberofstarsinbox xl=", xl, " xu=", xu, " yl=", yl, " yu=", yu, &
!   " noofstars=", numberofstarsinbox


end Function NumberOfStarsInBox

!********************************************************
!
! Function CountStars
! 
!
!
!*******************************************************
logical function CountStars(theSource, source, i, cosTheta, sinTheta, useStar, &
               minRadius, maxRadius, numStars, minimumCountRate,  &
                maxDistance)
    type(sourceparameters), intent(in) :: theSource

    type(sourceparameters), pointer :: source(:)

    integer(int16), intent(in) :: i
  
    real(single), intent(in) :: cosTheta, sinTheta
    logical, pointer :: useStar(:)
    integer(int16), intent(inout) :: minRadius, maxRadius
    integer(int16), intent(inout) :: numStars 
    real(single), intent(in) :: minimumCountRate, maxDistance
    integer(int16) :: j, k, jk1
    real(single) :: xpos, yPos
    real(single) :: radius
   
  
  
    real(single) :: dx, dy
 
 
    integer(int16) :: count  
    integer(int32) :: il, iu, n
   
 

    count = 0 ! The number of possible stars along the diffraction spike
       
    
    n = size(source)

!    write(*, *) "Countstars minradius=", minradius, " max=", maxradius, &
!      " xpos=", theSource%xpos, " y=", theSource%ypos
 
    jk1 = minradius + 30
    do j = minRadius, maxRadius             ! The radii from star to be checked

       
        radius = real(j)

        xpos = theSource%xpos + radius * costheta      
        yPos = theSource%ypos + radius * sintheta      ! Y-position along possible spike

        if(xPos .lt. 0.0 .or. ypos < 0.0 .or. xpos > 2048.5 .or. ypos > 2048.5) then



            if(count > 0) then
                numStars = numStars + count
                countStars = .true.
            else
                countStars = .false.
            end if

            maxRadius = j
            return

        end if
      
        il = 1
        il = locate(n, source%xpos, max(0.0, xpos - maxDistance - 2.0), il)
           
        iu = il
        iu = locate(n, source%xpos, xpos + maxDistance + 2.0, iu)

        do k = il, iu

            if (.not. usestar(k)) cycle             
            if (k .eq. i) cycle

            dx = abs(xpos - source(k)%xpos)
            if (dx .gt. maxDistance) cycle 
     
            dy = abs(ypos - source(k)%ypos)
            if (dy .gt. maxDistance) cycle 
     
            radius = sqrt(dx * dx + dy * dy)

            if (radius .gt. maxDistance) cycle 

            
            if ( (source(k)%rawCountRate > minimumCountRate .and. source(k)%rawCountRate < 999.0) &
                     .and. .not. isNullValue(source(k)%rawCountRate)) then
                useStar(k) = .false.
                cycle
            end if 
                          
            usestar(k) = .false.  ! Remember we have used this star         
            count = count + 1     ! The total number of stars along the possible spike
            jk1 = j
            
        end do

        if ( (j - jk1) > 60) exit            
    end do


!write(*, *) "Count=", count, " numstars=", numstars

    
    numStars = numStars + count

    if(count > 0) then
        countStars = .true.
    else
        countStars = .false.
    end if

end Function countStars

!********************************************************
!
! Function CheckForStars
! 
!
!
!*******************************************************
logical Function CheckForStars(theSource, source, i, theta, useStar, &
               minRad, maxRad, count3, minimumCountRate)
    type(sourceparameters), intent(in) :: theSource

    type(sourceparameters), pointer :: source(:)

    real(single), intent(in) :: theta
    integer(int16), intent(in) :: i
  

    logical, pointer :: useStar(:)
    integer(int16), intent(inout) :: minRad, maxRad
    integer(int16), intent(inout) :: count3  
    real(single), intent(in) :: minimumCountRate
    integer(int16) :: j, k
    real(single) :: xpos, yPos, theta1, theta2
    real(single) :: radius, costheta, sintheta
    real(single), parameter :: maxdistance = 5.0
    real(single), parameter :: maxDeltatheta = 25.0 
  
    real(single) :: dx, dy
    integer(int16),  parameter :: minRadius = 10, maxRadius = 120  ! Maximum radial distance from star to check
 
    integer(int16) :: count1, count2  
    integer(int32) :: il, iu, n
   
    theta1 = theta * degreesToRadians
  
    costheta = cos(theta1)
    sintheta = sin(theta1)

    count1 = 0
    count2 = 0
    theta1 = mod(real(theta), 180.0)
     

    checkForStars = .false.
    minrad = minRadius       
    maxrad = minradius + 30

    n = size(source)

  

!    write(*, *) "chekforstars i=", i, " theta=", theta

    usestar = .true.
    do j = minRadius, maxRadius             ! The radii from star to be checked

       
        radius = real(j)
        xpos = theSource%xpos + radius * costheta      
        yPos = theSource%ypos + radius * sintheta      ! Y-position along possible spike

        il = 1
        il = locate(n, source%xpos, max(0.0, xpos - maxDistance - 2.0), il)
           
        iu = il
        iu = locate(n, source%xpos, xpos + maxDistance + 2.0, iu)

!write(*, *) "Radius ", j, " il=", il, " iu=", iu            


        do k = il, iu

            if (.not. usestar(k)) cycle             
            if (k .eq. i) cycle

            dx = abs(xpos - source(k)%xpos)
            if (dx .gt. maxDistance) cycle 
     
            dy = abs(ypos - source(k)%ypos)
            if (dy .gt. maxDistance) cycle 
     
            radius = sqrt(dx * dx + dy * dy)

!if (radius <= maxdistance) then
!write(*, *) k, ") dx=", dx, " dy=", dy, " radius=", radius, " usestar=", usestar(k) 
!end if
            if (radius .gt. maxDistance) cycle 

            
            if ( (source(k)%rawCountRate > minimumCountRate .and. source(k)%rawCountRate < 999.0) &
                     .and. .not. isNullValue(source(k)%rawCountRate)) then
                checkForStars = .false.                             
                return
            end if 


!write(*, *) k, ") x=", source(k)%xpos, " y=", source(k)%ypos, " theta=", source(k)%theta

                          
            usestar(k) = .false.  ! Remember we have used this star         
            count2 = count2 + 1   ! The total number of stars along the possible spike
                                                        
            !*************************************************** 
            ! Check that the source is elongated along the spike
            !***************************************************
      
            theta2 = mod(source(k)%theta, 180.0)
           
 
            if ( abs(theta1 - theta2) .le. maxDeltaTheta) then ! Possible source aligned along spike
                  
                maxRad = j
                count1 = count1 + 1
             
!                write(*, *) "Line 1826 - possible star along spike  star no ", k, " theta1=", theta1, " theta2=", theta2
!                write(*, *) "count1=", count1, " count2=", count2, " j=", j, " maxrad=", maxrad
!                write(*, *) "theta2=", source(k)%theta                    
            end if

            
        end do

        if ( count2 > 10 .or. count1 > 4) exit

   
        if (count1 > 0 .and. (j - maxRad) .gt. 30) then
            exit
        elseif((j - maxrad) > 30) then
            exit
        end if
        
        
       
    end do

!write(*, *) "checkforstars count1=", count1, " count2=", count2, " count3=", count3


    if (count2 > 0) then

        if ( ( real(count1) / real(count2)  >= 0.5 ) .or. (count2 > 0 .and. count2 < 5)) then    
            CheckForStars = .true.
            count3 = count3 + count1
!write(*, *) "Checkforstars=true"

        else
            CheckForStars = .false.
           
        end if

    else
        CheckForStars = .false.
           
    end if

  
end Function CheckForStars

!********************************************************
!
!
!
!
!*******************************************************
subroutine SetStarSpikeFlags(xc, yc, source, index, il, iu, costheta, sintheta, minRadius, &
         maxRadius, usestar) 
   
    real(single), intent(in) :: xc, yc                         
    type(sourceparameters), pointer :: source(:)
    integer(int16), intent(in) :: index
    integer(int32), intent(in) :: il, iu
    real(single), intent(in) :: costheta, sintheta
  
    integer(int16), intent(in) :: minRadius, maxRadius
    logical, pointer :: useStar(:)
    integer(int16) :: i, j, k
    real(single) :: xpos, yPos
    real(single) :: radius
    real(single) :: tolerance
    integer(int16), parameter :: deltaik = 80 ! was 80
    integer(int16) :: count1
     
    real(single) :: dx, dy
    tolerance = 10.0

      
 
!write(*, *) "Setstarspikeflags minradius=", minradius, " maxradius=", maxradius, &
!  " theta=", theta, " xc=", xc, " yc=", yc
    k = minRadius

!write(*, *) "il=", il, " iu=", iu

    count1 = 0
    do i = minRadius, maxRadius  ! The radii from star to be checked
        radius = real(i)
        xpos = xc + radius * costheta      
        yPos = yc + radius * sintheta      

        do j = il, iu
        
            if (j .eq. index) cycle
            if (.not. useStar(j)) cycle 
               

            dx = abs(xpos - source(j)%xpos)
            if (dx > tolerance) cycle  
            dy = abs(ypos - source(j)%ypos)

            if (dy > tolerance) cycle  
            radius = sqrt(dx * dx + dy * dy)
          
            if (radius .le. tolerance) then
            
                count1 = count1 + 1
       

                source(j)%qualityFlag = ibset(source(j)%qualityFlag, 3)
!write(*, *) j, ") set star flag x=", source(j)%xpos, " y=", source(j)%ypos
                k = i     
                usestar(j) = .false.                
            end if
        end do
        if (count1 == 0 .and. i > 80) exit
        if ( (i - k) .gt. deltaik .and. count1 > 0) exit
        if (count1 > 20) exit        
    end do

!write(*, *) "No of flags set=", count1

end subroutine SetStarSpikeFlags


end module omsourceflags_mod









